# coding=utf-8
# Copyright 2023 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# modified here to be less messy, but still has problems

import random
import datetime
import numpy as np
import pandas as pd
from autots.models.base import ModelObject, PredictionObject
from autots.tools.probabilistic import Point_to_Probability

try:
    from sklearn.preprocessing import StandardScaler
    import tensorflow as tf
    from tensorflow import keras
    from pandas.tseries.holiday import EasterMonday
    from pandas.tseries.holiday import GoodFriday
    from pandas.tseries.holiday import Holiday
    from pandas.tseries.holiday import SU
    from pandas.tseries.holiday import TH
    from pandas.tseries.holiday import USColumbusDay
    from pandas.tseries.holiday import USLaborDay
    from pandas.tseries.holiday import USMartinLutherKingJr
    from pandas.tseries.holiday import USMemorialDay
    from pandas.tseries.holiday import USPresidentsDay
    from pandas.tseries.holiday import USThanksgivingDay
    from pandas.tseries.offsets import DateOffset
    from pandas.tseries.offsets import Day
    from pandas.tseries.offsets import Easter
    from tqdm import tqdm

    full_import = True
except Exception:
    full_import = False


def _distance_to_holiday(holiday):
    """Return distance to given holiday."""

    def _distance_to_day(index, MAX_WINDOW=200):
        # This is 183 to cover half a year (in both directions), also for leap years
        # + 17 as Eastern can be between March, 22 - April, 25
        holiday_date = holiday.dates(
            index - pd.Timedelta(days=MAX_WINDOW),
            index + pd.Timedelta(days=MAX_WINDOW),
        )
        assert (
            len(holiday_date) != 0  # pylint: disable=g-explicit-length-test
        ), f"No closest holiday for the date index {index} found."
        # It sometimes returns two dates if it is exactly half a year after the
        # holiday. In this case, the smaller distance (182 days) is returned.
        return (index - holiday_date[0]).days

    return _distance_to_day


def get_HOLIDAYS():
    EasterSunday = Holiday("Easter Sunday", month=1, day=1, offset=[Easter(), Day(0)])
    NewYearsDay = Holiday("New Years Day", month=1, day=1)
    SuperBowl = Holiday("Superbowl", month=2, day=1, offset=DateOffset(weekday=SU(1)))
    MothersDay = Holiday(
        "Mothers Day", month=5, day=1, offset=DateOffset(weekday=SU(2))
    )
    IndependenceDay = Holiday("Independence Day", month=7, day=4)
    ChristmasEve = Holiday("Christmas", month=12, day=24)
    ChristmasDay = Holiday("Christmas", month=12, day=25)
    NewYearsEve = Holiday("New Years Eve", month=12, day=31)
    BlackFriday = Holiday(
        "Black Friday",
        month=11,
        day=1,
        offset=[pd.DateOffset(weekday=TH(4)), Day(1)],
    )
    CyberMonday = Holiday(
        "Cyber Monday",
        month=11,
        day=1,
        offset=[pd.DateOffset(weekday=TH(4)), Day(4)],
    )

    HOLIDAYS = [
        EasterMonday,
        GoodFriday,
        USColumbusDay,
        USLaborDay,
        USMartinLutherKingJr,
        USMemorialDay,
        USPresidentsDay,
        USThanksgivingDay,
        EasterSunday,
        NewYearsDay,
        SuperBowl,
        MothersDay,
        IndependenceDay,
        ChristmasEve,
        ChristmasDay,
        NewYearsEve,
        BlackFriday,
        CyberMonday,
    ]
    return HOLIDAYS


class TimeCovariates(object):
    """Extract all time covariates except for holidays."""

    def __init__(
        self,
        datetimes,
        normalized=True,
        holiday=False,
    ):
        """Init function.

        Args:
          datetimes: pandas DatetimeIndex (lowest granularity supported is min)
          normalized: whether to normalize features or not
          holiday: fetch holiday features or not

        Returns:
          None
        """
        self.normalized = normalized
        self.dti = datetimes
        self.holiday = holiday

    def _minute_of_hour(self):
        minutes = np.array(self.dti.minute, dtype=np.float32)
        if self.normalized:
            minutes = minutes / 59.0 - 0.5
        return minutes

    def _hour_of_day(self):
        hours = np.array(self.dti.hour, dtype=np.float32)
        if self.normalized:
            hours = hours / 23.0 - 0.5
        return hours

    def _day_of_week(self):
        day_week = np.array(self.dti.dayofweek, dtype=np.float32)
        if self.normalized:
            day_week = day_week / 6.0 - 0.5
        return day_week

    def _day_of_month(self):
        day_month = np.array(self.dti.day, dtype=np.float32)
        if self.normalized:
            day_month = day_month / 30.0 - 0.5
        return day_month

    def _day_of_year(self):
        day_year = np.array(self.dti.dayofyear, dtype=np.float32)
        if self.normalized:
            day_year = day_year / 364.0 - 0.5
        return day_year

    def _month_of_year(self):
        month_year = np.array(self.dti.month, dtype=np.float32)
        if self.normalized:
            month_year = month_year / 11.0 - 0.5
        return month_year

    def _week_of_year(self):
        week_year = np.array(self.dti.strftime("%U").astype(int), dtype=np.float32)
        if self.normalized:
            week_year = week_year / 51.0 - 0.5
        return week_year

    def _get_holidays(self):
        dti_series = self.dti.to_series()
        HOLIDAYS = get_HOLIDAYS()
        hol_variates = np.vstack(
            [dti_series.apply(_distance_to_holiday(h)).values for h in tqdm(HOLIDAYS)]
        )
        # hol_variates is (num_holiday, num_time_steps), the normalization should be
        # performed in the num_time_steps dimension.
        return StandardScaler().fit_transform(hol_variates.T).T

    def get_covariates(self):
        """Get all time covariates."""
        moh = self._minute_of_hour().reshape(1, -1)
        hod = self._hour_of_day().reshape(1, -1)
        dom = self._day_of_month().reshape(1, -1)
        dow = self._day_of_week().reshape(1, -1)
        doy = self._day_of_year().reshape(1, -1)
        moy = self._month_of_year().reshape(1, -1)
        woy = self._week_of_year().reshape(1, -1)

        all_covs = [
            moh,
            hod,
            dom,
            dow,
            doy,
            moy,
            woy,
        ]
        columns = ["moh", "hod", "dom", "dow", "doy", "moy", "woy"]
        if self.holiday:
            hol_covs = self._get_holidays()
            HOLIDAYS = get_HOLIDAYS()
            all_covs.append(hol_covs)
            columns += [f"hol_{i}" for i in range(len(HOLIDAYS))]

        return pd.DataFrame(
            data=np.vstack(all_covs).transpose(),
            columns=columns,
            index=self.dti,
        )


class TimeSeriesdata(object):
    """Data loader class."""

    def __init__(
        self,
        df,
        num_cov_cols,
        cat_cov_cols,
        ts_cols,
        train_range,
        val_range,
        test_range,
        hist_len,
        pred_len,
        batch_size,
        freq="D",
        normalize=True,
        epoch_len=None,
        holiday=False,
        permute=True,
    ):
        """Initialize objects.

        Args:
          df: wide style dataframe
          num_cov_cols: list of numerical global covariates
          cat_cov_cols: list of categorical global covariates
          ts_cols: columns corresponding to ts
          train_range: tuple of train ranges
          val_range: tuple of validation ranges
          test_range: tuple of test ranges
          hist_len: historical context
          pred_len: prediction length
          batch_size: batch size (number of ts in a batch)
          freq: freq of original data
          normalize: std. normalize data or not
          epoch_len: num iters in an epoch
          holiday: use holiday features or not
          permute: permute ts in train batches or not

        Returns:
          None
        """
        self.data_df = df
        if not num_cov_cols:
            self.data_df["ncol"] = np.zeros(self.data_df.shape[0])
            num_cov_cols = ["ncol"]
        if not cat_cov_cols:
            self.data_df["ccol"] = np.zeros(self.data_df.shape[0])
            cat_cov_cols = ["ccol"]
        self.data_df.fillna(0, inplace=True)
        # self.data_df.set_index(pd.DatetimeIndex(self.data_df[datetime_col]), inplace=True)
        self.num_cov_cols = num_cov_cols
        self.cat_cov_cols = cat_cov_cols
        self.ts_cols = ts_cols
        self.train_range = train_range
        self.val_range = val_range
        self.test_range = test_range
        data_df_idx = self.data_df.index
        date_index = data_df_idx.union(
            pd.date_range(
                data_df_idx[-1] + pd.Timedelta(1, freq=freq),
                periods=pred_len + 1,
                freq=freq,
            )
        )
        self.time_df = TimeCovariates(date_index, holiday=holiday).get_covariates()
        self.hist_len = hist_len
        self.pred_len = pred_len
        self.batch_size = batch_size
        self.freq = freq
        self.normalize = normalize
        self.data_mat = self.data_df[self.ts_cols].to_numpy().transpose()
        self.data_mat = self.data_mat[:, 0 : self.test_range[1]]
        self.time_mat = self.time_df.to_numpy().transpose()
        self.num_feat_mat = self.data_df[num_cov_cols].to_numpy().transpose()
        self.cat_feat_mat, self.cat_sizes = self._get_cat_cols(cat_cov_cols)
        self.normalize = normalize
        if normalize:
            self._normalize_data()
        self.epoch_len = epoch_len
        self.permute = permute

    def _get_cat_cols(self, cat_cov_cols):
        """Get categorical columns."""
        cat_vars = []
        cat_sizes = []
        for col in cat_cov_cols:
            dct = {x: i for i, x in enumerate(self.data_df[col].unique())}
            cat_sizes.append(len(dct))
            mapped = (
                self.data_df[col].map(lambda x: dct[x]).to_numpy().transpose()
            )  # pylint: disable=cell-var-from-loop
            cat_vars.append(mapped)
        return np.vstack(cat_vars), cat_sizes

    def _normalize_data(self):
        self.scaler = StandardScaler()
        train_mat = self.data_mat[:, self.train_range[0] : self.train_range[1]]
        self.scaler = self.scaler.fit(train_mat.transpose())
        self.data_mat = self.scaler.transform(self.data_mat.transpose()).transpose()

    def train_gen(self):
        """Generator for training data."""
        num_ts = len(self.ts_cols)
        perm = np.arange(
            self.train_range[0] + self.hist_len,
            self.train_range[1] - self.pred_len,
        )
        perm = np.random.permutation(perm)
        hist_len = self.hist_len
        if not self.epoch_len:
            epoch_len = len(perm)
        else:
            epoch_len = self.epoch_len
        for idx in perm[0:epoch_len]:
            for _ in range(num_ts // self.batch_size + 1):
                if self.permute:
                    tsidx = np.random.choice(
                        num_ts, size=self.batch_size, replace=False
                    )
                else:
                    tsidx = np.arange(num_ts)
                dtimes = np.arange(idx - hist_len, idx + self.pred_len)
                (
                    bts_train,
                    bts_pred,
                    bfeats_train,
                    bfeats_pred,
                    bcf_train,
                    bcf_pred,
                ) = self._get_features_and_ts(dtimes, tsidx, hist_len)

                all_data = [
                    bts_train,
                    bfeats_train,
                    bcf_train,
                    bts_pred,
                    bfeats_pred,
                    bcf_pred,
                    tsidx,
                ]
                yield tuple(all_data)

    def test_val_gen(self, mode="val"):
        """Generator for validation/test data."""
        if mode == "val":
            start = self.val_range[0]
            end = self.val_range[1] - self.pred_len + 1
        elif mode == "test":
            start = self.test_range[0]
            end = self.test_range[1] - self.pred_len + 1
        elif mode == "predict":
            start = self.test_range[0]
            end = self.test_range[1] + 1
        else:
            raise NotImplementedError("Eval mode not implemented")
        num_ts = len(self.ts_cols)
        hist_len = self.hist_len
        perm = np.arange(start, end)
        if self.epoch_len:
            epoch_len = self.epoch_len
        else:
            epoch_len = len(perm)
        for idx in perm[0:epoch_len]:
            for batch_idx in range(0, num_ts, self.batch_size):
                tsidx = np.arange(batch_idx, min(batch_idx + self.batch_size, num_ts))
                dtimes = np.arange(idx - hist_len, idx + self.pred_len)
                (
                    bts_train,
                    bts_pred,
                    bfeats_train,
                    bfeats_pred,
                    bcf_train,
                    bcf_pred,
                ) = self._get_features_and_ts(dtimes, tsidx, hist_len)
                all_data = [
                    bts_train,
                    bfeats_train,
                    bcf_train,
                    bts_pred,
                    bfeats_pred,
                    bcf_pred,
                    tsidx,
                ]
                yield tuple(all_data)

    def _get_features_and_ts(self, dtimes, tsidx, hist_len=None):
        """Get features and ts in specified windows."""
        if hist_len is None:
            hist_len = self.hist_len
        data_times = dtimes[dtimes < self.data_mat.shape[1]]
        bdata = self.data_mat[:, data_times]
        bts = bdata[tsidx, :]
        bnf = self.num_feat_mat[:, data_times]
        bcf = self.cat_feat_mat[:, data_times]
        btf = self.time_mat[:, dtimes]
        if bnf.shape[1] < btf.shape[1]:
            rem_len = btf.shape[1] - bnf.shape[1]
            rem_rep = np.repeat(bnf[:, [-1]], repeats=rem_len)
            rem_rep_cat = np.repeat(bcf[:, [-1]], repeats=rem_len)
            bnf = np.hstack([bnf, rem_rep.reshape(bnf.shape[0], -1)])
            bcf = np.hstack([bcf, rem_rep_cat.reshape(bcf.shape[0], -1)])
        bfeats = np.vstack([btf, bnf])
        bts_train = bts[:, 0:hist_len]
        bts_pred = bts[:, hist_len:]
        bfeats_train = bfeats[:, 0:hist_len]
        bfeats_pred = bfeats[:, hist_len:]
        bcf_train = bcf[:, 0:hist_len]
        bcf_pred = bcf[:, hist_len:]
        return bts_train, bts_pred, bfeats_train, bfeats_pred, bcf_train, bcf_pred

    def tf_dataset(self, mode="train"):
        """Tensorflow Dataset."""
        if mode == "train":
            gen_fn = self.train_gen
        else:
            gen_fn = lambda: self.test_val_gen(mode)
        output_types = tuple(
            [tf.float32] * 2 + [tf.int32] + [tf.float32] * 2 + [tf.int32] * 2
        )
        dataset = tf.data.Dataset.from_generator(gen_fn, output_types)
        # dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
        return dataset


EPS = 1e-7


if full_import:

    class MLPResidual(keras.layers.Layer):
        """Simple one hidden state residual network."""

        def __init__(self, hidden_dim, output_dim, layer_norm=False, dropout_rate=0.0):
            super(MLPResidual, self).__init__()
            self.lin_a = tf.keras.layers.Dense(
                hidden_dim,
                activation="relu",
            )
            self.lin_b = tf.keras.layers.Dense(
                output_dim,
                activation=None,
            )
            self.lin_res = tf.keras.layers.Dense(
                output_dim,
                activation=None,
            )
            if layer_norm:
                self.lnorm = tf.keras.layers.LayerNormalization()
            self.layer_norm = layer_norm
            self.dropout = tf.keras.layers.Dropout(dropout_rate)

        def call(self, inputs):
            """Call method."""
            h_state = self.lin_a(inputs)
            out = self.lin_b(h_state)
            out = self.dropout(out)
            res = self.lin_res(inputs)
            if self.layer_norm:
                return self.lnorm(out + res)
            return out + res

    def _make_dnn_residual(hidden_dims, layer_norm=False, dropout_rate=0.0):
        """Multi-layer DNN residual model."""
        if len(hidden_dims) < 2:
            return keras.layers.Dense(
                hidden_dims[-1],
                activation=None,
            )
        layers = []
        for i, hdim in enumerate(hidden_dims[:-1]):
            layers.append(
                MLPResidual(
                    hdim,
                    hidden_dims[i + 1],
                    layer_norm=layer_norm,
                    dropout_rate=dropout_rate,
                )
            )
        return keras.Sequential(layers)

    class TideModel(keras.Model):
        """Main class for multi-scale DNN model."""

        def __init__(
            self,
            model_config,
            pred_len,
            cat_sizes,
            num_ts,
            transform=False,
            cat_emb_size=4,
            layer_norm=False,
            dropout_rate=0.0,
        ):
            """Tide model.

            Args:
              model_config: configurations specific to the model.
              pred_len: prediction horizon length.
              cat_sizes: number of categories in each categorical covariate.
              num_ts: number of time-series in the dataset
              transform: apply reversible transform or not.
              cat_emb_size: embedding size of categorical variables.
              layer_norm: use layer norm or not.
              dropout_rate: level of dropout.
            """
            super().__init__()
            self.model_config = model_config
            self.transform = transform
            if self.transform:
                self.affine_weight = self.add_weight(
                    name="affine_weight",
                    shape=(num_ts,),
                    initializer="ones",
                    trainable=True,
                )

                self.affine_bias = self.add_weight(
                    name="affine_bias",
                    shape=(num_ts,),
                    initializer="zeros",
                    trainable=True,
                )
            self.pred_len = pred_len
            self.encoder = _make_dnn_residual(
                model_config.get("hidden_dims"),
                layer_norm=layer_norm,
                dropout_rate=dropout_rate,
            )
            self.decoder = _make_dnn_residual(
                model_config.get("hidden_dims")[:-1]
                + [
                    model_config.get("decoder_output_dim") * self.pred_len,
                ],
                layer_norm=layer_norm,
                dropout_rate=dropout_rate,
            )
            self.linear = tf.keras.layers.Dense(
                self.pred_len,
                activation=None,
            )
            self.time_encoder = _make_dnn_residual(
                model_config.get("time_encoder_dims"),
                layer_norm=layer_norm,
                dropout_rate=dropout_rate,
            )
            self.final_decoder = MLPResidual(
                hidden_dim=model_config.get("final_decoder_hidden"),
                output_dim=1,
                layer_norm=layer_norm,
                dropout_rate=dropout_rate,
            )
            self.cat_embs = []
            for cat_size in cat_sizes:
                self.cat_embs.append(
                    tf.keras.layers.Embedding(
                        input_dim=cat_size, output_dim=cat_emb_size
                    )
                )
            self.ts_embs = tf.keras.layers.Embedding(input_dim=num_ts, output_dim=16)
            self.train_loss = keras.losses.MeanSquaredError()

        @tf.function
        def _assemble_feats(self, feats, cfeats):
            """assemble all features."""
            all_feats = [feats]
            for i, emb in enumerate(self.cat_embs):
                all_feats.append(tf.transpose(emb(cfeats[i, :])))
            return tf.concat(all_feats, axis=0)

        @tf.function
        def call(self, inputs):
            """Call function that takes in a batch of training data and features."""
            past_data = inputs[0]
            future_features = inputs[1]
            bsize = past_data[0].shape[0]
            tsidx = inputs[2]
            past_feats = self._assemble_feats(past_data[1], past_data[2])
            future_feats = self._assemble_feats(future_features[0], future_features[1])
            past_ts = past_data[0]
            if self.transform:
                affine_weight = tf.gather(self.affine_weight, tsidx)
                affine_bias = tf.gather(self.affine_bias, tsidx)
                batch_mean = tf.math.reduce_mean(past_ts, axis=1)
                batch_std = tf.math.reduce_std(past_ts, axis=1)
                batch_std = tf.where(
                    tf.math.equal(batch_std, 0.0), tf.ones_like(batch_std), batch_std
                )
                past_ts = (past_ts - batch_mean[:, None]) / batch_std[:, None]
                past_ts = affine_weight[:, None] * past_ts + affine_bias[:, None]
            encoded_past_feats = tf.transpose(
                self.time_encoder(tf.transpose(past_feats))
            )
            encoded_future_feats = tf.transpose(
                self.time_encoder(tf.transpose(future_feats))
            )
            enc_past = tf.repeat(tf.expand_dims(encoded_past_feats, axis=0), bsize, 0)
            enc_past = tf.reshape(enc_past, [bsize, -1])
            enc_fut = tf.repeat(
                tf.expand_dims(encoded_future_feats, axis=0), bsize, 0
            )  # batch x fdim x H
            enc_future = tf.reshape(enc_fut, [bsize, -1])
            residual_out = self.linear(past_ts)
            ts_embs = self.ts_embs(tsidx)
            encoder_input = tf.concat([past_ts, enc_past, enc_future, ts_embs], axis=1)
            encoding = self.encoder(encoder_input)
            decoder_out = self.decoder(encoding)
            decoder_out = tf.reshape(
                decoder_out, [bsize, -1, self.pred_len]
            )  # batch x d x H
            final_in = tf.concat([decoder_out, enc_fut], axis=1)
            out = self.final_decoder(tf.transpose(final_in, (0, 2, 1)))  # B x H x 1
            out = tf.squeeze(out, axis=-1)
            out += residual_out
            if self.transform:
                out = (out - affine_bias[:, None]) / (affine_weight[:, None] + EPS)
                out = out * batch_std[:, None] + batch_mean[:, None]
            return out

        @tf.function
        def train_step(self, past_data, future_features, ytrue, tsidx, optimizer):
            """One step of training."""
            with tf.GradientTape() as tape:
                all_preds = self((past_data, future_features, tsidx), training=True)
                loss = self.train_loss(ytrue, all_preds)

            grads = tape.gradient(loss, self.trainable_variables)
            optimizer.apply_gradients(zip(grads, self.trainable_variables))
            return loss

        def get_all_eval_data(self, data, mode, num_split=1):
            y_preds = []
            y_trues = []
            all_test_loss = 0
            all_test_num = 0
            idxs = np.arange(0, self.pred_len, self.pred_len // num_split).tolist() + [
                self.pred_len
            ]
            for i in range(len(idxs) - 1):
                indices = (idxs[i], idxs[i + 1])
                (
                    all_y_true,
                    all_y_pred,
                    test_loss,
                    test_num,
                ) = self.get_eval_data_for_split(data, mode, indices)
                y_preds.append(all_y_pred)
                y_trues.append(all_y_true)
                all_test_loss += test_loss
                all_test_num += test_num
            return np.hstack(y_preds), np.hstack(y_trues), all_test_loss / all_test_num

        def get_eval_data_for_split(self, data, mode, indices):
            iterator = data.tf_dataset(mode=mode)

            all_y_true = None
            all_y_pred = None

            def set_or_concat(a, b):
                if a is None:
                    return b
                return tf.concat((a, b), axis=1)

            all_test_loss = 0
            all_test_num = 0
            ts_count = 0
            ypreds = []
            ytrues = []
            for all_data in tqdm(iterator):
                past_data = all_data[:3]
                future_features = all_data[4:6]
                y_true = all_data[3]
                tsidx = all_data[-1]
                all_preds = self((past_data, future_features, tsidx), training=False)
                y_pred = all_preds
                y_pred = y_pred[:, 0 : y_true.shape[1]]
                id1 = indices[0]
                id2 = min(indices[1], y_true.shape[1])
                y_pred = y_pred[:, id1:id2]
                y_true = y_true[:, id1:id2]
                loss = self.train_loss(y_true, y_pred)
                all_test_loss += loss
                all_test_num += 1
                ts_count += y_true.shape[0]
                ypreds.append(y_pred)
                ytrues.append(y_true)
                if ts_count >= len(data.ts_cols):
                    ts_count = 0
                    ypreds = tf.concat(ypreds, axis=0)
                    ytrues = tf.concat(ytrues, axis=0)
                    all_y_true = set_or_concat(all_y_true, ytrues)
                    all_y_pred = set_or_concat(all_y_pred, ypreds)
                    ypreds = []
                    ytrues = []
            return (
                all_y_true.numpy(),
                all_y_pred.numpy(),
                all_test_loss.numpy(),
                all_test_num,
            )

        def evaluate(self, data, mode, num_split=1):
            all_y_pred, all_y_true, test_loss = self.get_all_eval_data(
                data, mode, num_split
            )

            result_dict = {}
            for metric in METRICS:
                eval_fn = METRICS[metric]
                result_dict[metric] = np.float64(eval_fn(all_y_pred, all_y_true))

            return (
                result_dict,
                (all_y_pred, all_y_true),
                test_loss,
            )


def mape(y_pred, y_true):
    abs_diff = np.abs(y_pred - y_true).flatten()
    abs_val = np.abs(y_true).flatten()
    idx = np.where(abs_val > EPS)
    mpe = np.mean(abs_diff[idx] / abs_val[idx])
    return mpe


def mae_loss(y_pred, y_true):
    return np.abs(y_pred - y_true).mean()


def wape(y_pred, y_true):
    abs_diff = np.abs(y_pred - y_true)
    abs_val = np.abs(y_true)
    wpe = np.sum(abs_diff) / (np.sum(abs_val) + EPS)
    return wpe


def smape(y_pred, y_true):
    abs_diff = np.abs(y_pred - y_true)
    abs_mean = (np.abs(y_true) + np.abs(y_pred)) / 2
    smpe = np.mean(abs_diff / (abs_mean + EPS))
    return smpe


def rmse(y_pred, y_true):
    return np.sqrt(np.square(y_pred - y_true).mean())


def nrmse(y_pred, y_true):
    mse = np.square(y_pred - y_true)
    return np.sqrt(mse.mean()) / np.abs(y_true).mean()


METRICS = {
    "mape": mape,
    "wape": wape,
    "smape": smape,
    "nrmse": nrmse,
    "rmse": rmse,
    "mae": mae_loss,
}


class TiDE(ModelObject):
    """Google Research MLP based forecasting approach."""

    def __init__(
        self,
        name: str = "UnivariateRegression",
        random_seed=42,
        frequency="D",
        learning_rate=0.0009999,
        transform=False,
        layer_norm=False,
        holiday=True,
        dropout_rate=0.3,
        batch_size=512,
        hidden_size=256,
        num_layers=1,
        hist_len=720,
        decoder_output_dim=16,
        final_decoder_hidden=64,
        num_split=4,
        min_num_epochs=0,
        train_epochs=100,
        patience=10,
        epoch_len=None,
        permute=True,
        normalize=True,
        gpu_index=0,
        n_jobs: int = "auto",
        forecast_length: int = 14,
        prediction_interval: float = 0.9,
        verbose: int = 0,
    ):
        ModelObject.__init__(
            self,
            name="TiDE",
            frequency=frequency,
            prediction_interval=prediction_interval,
            regression_type=None,
            holiday_country=None,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
        )
        self.forecast_length = forecast_length
        self.learning_rate = learning_rate
        self.transform = transform
        self.layer_norm = layer_norm
        self.holiday = holiday
        self.dropout_rate = dropout_rate
        self.batch_size = batch_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.hist_len = hist_len
        self.decoder_output_dim = decoder_output_dim
        self.final_decoder_hidden = final_decoder_hidden
        self.num_split = num_split
        self.min_num_epochs = min_num_epochs
        self.train_epochs = train_epochs
        self.patience = patience
        self.epoch_len = epoch_len
        self.permute = permute
        self.normalize = normalize
        self.gpu_index = gpu_index

    def fit(self, df=None, num_cov_cols=None, cat_cov_cols=None, future_regressor=None):
        """Training TS code."""
        df = self.basic_profile(df)
        tf.random.set_seed(self.random_seed)
        np.random.seed(self.random_seed)
        keras.backend.clear_session()
        data_df = df.copy()

        try:
            gpus = tf.config.experimental.list_physical_devices("GPU")
            if False:
                tf.config.experimental.set_visible_devices(gpus[self.gpu_index], "GPU")
                if gpus:
                    try:
                        for gpu in gpus:
                            tf.config.experimental.set_memory_growth(gpu, True)
                    except RuntimeError as e:
                        print(e)
        except Exception as e:
            print("TF GPU init error:" + repr(e))

        # some weird way of passing validation indexes which is silly for prod
        full_len_idx = data_df.shape[0]
        # will be tied to num_splits
        boundaries = [
            full_len_idx - self.forecast_length * 3,
            full_len_idx - self.forecast_length * 2,
            full_len_idx - self.forecast_length,
        ]

        if num_cov_cols is None and cat_cov_cols is None:
            ts_cols = data_df.columns
        else:
            num_cov_cols_temp = (
                [] if not isinstance(num_cov_cols, list) else num_cov_cols
            )
            cat_cov_cols_temp = (
                [] if not isinstance(cat_cov_cols, list) else cat_cov_cols
            )
            ts_cols = [
                col
                for col in data_df.columns
                if col not in set(num_cov_cols_temp + cat_cov_cols_temp)
            ]
        dtl = TimeSeriesdata(
            df=data_df,
            num_cov_cols=num_cov_cols,
            cat_cov_cols=None,
            ts_cols=np.array(ts_cols),
            train_range=[0, boundaries[0]],
            val_range=[boundaries[0], boundaries[1]],
            test_range=[boundaries[1], boundaries[2]],
            hist_len=self.hist_len,
            pred_len=self.forecast_length,
            batch_size=min(self.batch_size, len(ts_cols)),
            freq=self.frequency,
            normalize=self.normalize,
            epoch_len=self.epoch_len,
            holiday=self.holiday,
            permute=self.permute,
        )

        # Create model
        model_config = {
            "model_type": "dnn",
            "hidden_dims": [self.hidden_size] * self.num_layers,
            "time_encoder_dims": [64, 4],
            "decoder_output_dim": self.decoder_output_dim,
            "final_decoder_hidden": self.final_decoder_hidden,
            "batch_size": dtl.batch_size,
        }
        self.model = TideModel(
            model_config=model_config,
            pred_len=self.forecast_length,
            num_ts=len(ts_cols),
            cat_sizes=dtl.cat_sizes,
            transform=self.transform,
            layer_norm=self.layer_norm,
            dropout_rate=self.dropout_rate,
        )

        step = tf.Variable(0)
        # LR scheduling
        lr_schedule = keras.optimizers.schedules.CosineDecay(
            initial_learning_rate=self.learning_rate,
            decay_steps=30 * dtl.train_range[1],
        )

        optimizer = keras.optimizers.Adam(learning_rate=lr_schedule, clipvalue=1e3)

        best_loss = np.inf
        pat = 0
        # best_check_path = None
        while step.numpy() < self.train_epochs + 1:
            ep = step.numpy()
            # sys.stdout.flush()

            iterator = dtl.tf_dataset(mode="train")
            for i, batch in enumerate(iterator):
                # print(f"i is {i}")
                past_data = batch[:3]
                future_features = batch[4:6]
                tsidx = batch[-1]
                loss = self.model.train_step(  # noqa
                    past_data, future_features, batch[3], tsidx, optimizer
                )
                # print(f"training epoch is {ep} with loss: {loss}")

            step.assign_add(1)
            # Test metrics
            val_metrics, val_res, val_loss = self.model.evaluate(
                dtl, "val", num_split=self.num_split
            )
            test_metrics, test_res, test_loss = self.model.evaluate(
                dtl, "test", num_split=self.num_split
            )
            tracked_loss = val_metrics["rmse"]
            if tracked_loss < best_loss and ep > self.min_num_epochs:
                best_loss = tracked_loss
                pat = 0
            else:
                pat += 1
                # early stopping
                if pat > self.patience:
                    break

            if self.verbose > 0:
                print(
                    f"epoch {ep} with mape {val_metrics['mape']} and rmse {val_metrics['rmse']}"
                )
        self.past_data = past_data
        self.future_features = future_features
        self.tsidx = tsidx
        self.dtl = dtl
        self.fit_runtime = datetime.datetime.now() - self.startTime

        return self

    def predict(
        self,
        forecast_length="self",
        future_regressor=None,
        mode="test",
        just_point_forecast=False,
    ):
        if forecast_length != "self":
            if forecast_length != self.forecast_length:
                raise ValueError("forecast_length must be set in TiDE init")
        predictStartTime = datetime.datetime.now()
        iterator = self.dtl.tf_dataset(mode=mode)
        *_, batch = iterator
        past_data = batch[:3]
        future_features = batch[4:6]
        tsidx = batch[-1]
        all_preds = self.model((past_data, future_features, tsidx), training=False)
        if self.normalize:
            arr = self.dtl.scaler.inverse_transform(all_preds.numpy().T)
        else:
            arr = all_preds.numpy().T
        fut_idx = pd.date_range(
            self.train_last_date, periods=self.forecast_length + 1, freq=self.frequency
        )[1:]
        forecast = pd.DataFrame(arr, columns=self.column_names, index=fut_idx)
        if just_point_forecast:
            return forecast
        else:
            upper_forecast, lower_forecast = Point_to_Probability(
                self.dtl.data_df[self.column_names],
                forecast,
                prediction_interval=self.prediction_interval,
                method="inferred_normal",
            )

            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=self.forecast_length,
                forecast_index=forecast.index,
                forecast_columns=self.column_names,
                lower_forecast=lower_forecast,
                forecast=forecast,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )
            return prediction
        return forecast

    def get_new_params(self, method: str = "random"):
        """Return dict of new parameters for parameter tuning."""
        return {
            # borrowed mostly from the defaults used for the paper results
            "learning_rate": random.choices(
                [0.00006558, 0.000999999, 0.000030127, 0.01, 0.000001],
                [0.3, 0.3, 0.3, 0.05, 0.05],
            )[0],
            "transform": random.choices([True, False], [0.3, 0.7])[0],
            "layer_norm": random.choices([True, False], [0.2, 0.8])[0],
            "holiday": random.choices([True, False], [0.3, 0.7])[0],
            "dropout_rate": random.choices([0.0, 0.3, 0.5, 0.7], [0.3, 0.3, 0.3, 0.05])[
                0
            ],
            "batch_size": random.choices([1024, 512, 257, 32], [0.05, 0.4, 0.1, 0.1])[
                0
            ],
            "hidden_size": random.choices([1024, 512, 256, 64], [0.1, 0.2, 0.4, 0.1])[
                0
            ],
            "num_layers": random.choices([1, 2, 3], [0.85, 0.1, 0.02])[0],
            "hist_len": random.choices([720, 10, 364, 21], [0.1, 0.1, 0.1, 0.1])[0],
            "decoder_output_dim": random.choices([16, 8, 4, 32], [0.4, 0.4, 0.4, 0.1])[
                0
            ],
            "final_decoder_hidden": random.choices(
                [128, 64, 32, 16], [0.1, 0.8, 0.1, 0.1]
            )[0],
            "num_split": random.choices([6, 4, 1, 2], [0.1, 0.8, 0.2, 0.2])[0],
            "min_num_epochs": random.choices([5, 10, 20], [0.85, 0.1, 0.1])[0],
            "train_epochs": random.choices([20, 40, 100, 500], [0.2, 0.4, 0.2, 0.05])[
                0
            ],
            # "patience": random.choices([5000, 1000, 50000], [0.85, 0.05, 0.1])[0],
            "epoch_len": random.choices(
                [1, 4, 20, 40, None], [0.3, 0.2, 0.1, 0.1, 0.3]
            )[0],
            "permute": random.choices([True, False], [0.3, 0.7])[0],
            "normalize": random.choices([True, False], [0.5, 0.5])[0],
        }

    def get_params(self):
        """Return dict of current parameters."""
        return {
            "learning_rate": self.learning_rate,
            "transform": self.transform,
            "layer_norm": self.layer_norm,
            "holiday": self.holiday,
            "dropout_rate": self.dropout_rate,
            "batch_size": self.batch_size,
            "hidden_size": self.hidden_size,
            "num_layers": self.num_layers,
            "hist_len": self.hist_len,
            "decoder_output_dim": self.decoder_output_dim,
            "final_decoder_hidden": self.final_decoder_hidden,
            "num_split": self.num_split,
            "min_num_epochs": self.min_num_epochs,
            "train_epochs": self.train_epochs,
            # "patience": self.patience,
            "epoch_len": self.epoch_len,
            "permute": self.permute,
            "normalize": self.normalize,
        }


"""
import matplotlib.pyplot as plt
from autots import load_daily

df = load_daily(long=False)
params = TiDE().get_new_params()
print(params)
mod = TiDE(forecast_length=60, **params)
mod = mod.fit(df=df)
forecast = mod.predict()
col = 'wiki_all'
combined = pd.concat([df.iloc[-100:], forecast.forecast], axis=0)
for col in df.columns:
    ax = combined[col].plot(title=col)
    ax.vlines(
        x=forecast.forecast.index[0],
        ls='-.',
        lw=1,
        colors='darkred',
        ymin=combined[col].min(),
        ymax=combined[col].max(),
    )
    plt.savefig(f"TiDE_{col}", dpi=300)
    plt.show()
"""



"""
Constraint generation functions
"""

import random
import numpy as np
import pandas as pd

from autots.tools.impute import FillNA


def constant_growth_rate(periods, final_growth):
    """Take a final target growth rate (ie 2 % over a year) and convert to a daily (etc) value."""
    # Convert final growth rate percentage to a growth factor
    final_growth_factor = 1 + final_growth

    # Calculate the daily growth factor required to achieve the final growth factor in the given days
    daily_growth_factor = final_growth_factor ** (1 / periods)

    # Generate an array of day indices
    day_indices = np.arange(1, periods + 1)

    # Calculate the cumulative growth factor for each day
    cumulative_growth_factors = daily_growth_factor**day_indices

    # Calculate the perceived growth rates relative to the starting value
    perceived_growth_rates = cumulative_growth_factors - 1
    return perceived_growth_rates


constraint_method_dict = {
    "quantile": 0.1,
    "stdev": 0.1,
    "stdev_min": 0.1,
    "last_window": 0.1,
    "slope": 0.1,
    "absolute": 0.1,
    "fixed": 0,
    "historic_growth": 0.1,
    "historic_diff": 0.05,
    "dampening": 0.1,
    "round": 0.05,
}


def constraint_new_params(method: str = "fast"):
    method_choice = random.choices(
        list(constraint_method_dict.keys()), list(constraint_method_dict.values())
    )[0]
    params = {
        "constraint_method": method_choice,
        "constraint_direction": random.choices(["upper", "lower"], [0.5, 0.5])[0],
        "constraint_regularization": random.choices(
            [1.0, 0.2, 0.5, 0.7, 0.9], [0.7, 0.05, 0.1, 0.05, 0.1]
        )[0],
    }
    if method_choice == "quantile":
        if params["constraint_direction"] == "upper":
            params["constraint_value"] = random.choices(
                [1.0, 0.5, 0.7, 0.9, 0.98], [0.2, 0.2, 0.1, 0.2, 0.1]
            )[0]
        else:
            params["constraint_value"] = random.choices(
                [0.01, 0.5, 0.1, 0.2, 0.02], [0.2, 0.2, 0.1, 0.2, 0.1]
            )[0]
    elif method_choice == "slope":
        if params["constraint_direction"] == "upper":
            params["constraint_value"] = random.choices(
                [
                    {
                        "slope": 0.02,
                        "window": 10,
                        "window_agg": "max",
                        "threshold": 0.01,
                    },
                    {
                        "slope": 0.05,
                        "window": 10,
                        "window_agg": "max",
                        "threshold": 0.01,
                    },
                    {
                        "slope": 0.1,
                        "window": 30,
                        "window_agg": "max",
                        "threshold": 0.01,
                    },
                    {
                        "slope": 0.2,
                        "window": 10,
                        "window_agg": "mean",
                        "threshold": 0.01,
                    },
                    {
                        "slope": 0.001,
                        "window": 10,
                        "window_agg": "max",
                        "threshold": 0.1,
                    },
                ],
                [0.2, 0.2, 0.1, 0.2, 0.1],
            )[0]
        else:
            params["constraint_value"] = random.choices(
                [
                    {
                        "slope": -0.02,
                        "window": 7,
                        "window_agg": "min",
                        "threshold": -0.01,
                    },
                    {
                        "slope": -0.05,
                        "window": 10,
                        "window_agg": "min",
                        "threshold": -0.01,
                    },
                    {
                        "slope": -0.1,
                        "window": 30,
                        "window_agg": "min",
                        "threshold": 0.01,
                    },
                    {
                        "slope": -0.2,
                        "window": 10,
                        "window_agg": "mean",
                        "threshold": -0.01,
                    },
                    {
                        "slope": -0.001,
                        "window": 10,
                        "window_agg": "min",
                        "threshold": -0.1,
                    },
                ],
                [0.5, 0.2, 0.1, 0.2, 0.1],
            )[0]
    elif method_choice == "last_window":
        if params["constraint_direction"] == "upper":
            params["constraint_value"] = random.choices(
                [
                    {"window": 10, "threshold": 0.1},
                    {"window": 10, "threshold": 0.2},
                    {"window": 20, "threshold": 0.1},
                    {"window": 364, "threshold": 0.01},
                    {"window": 10, "threshold": -0.01},
                ],
                [0.5, 0.2, 0.1, 0.2, 0.1],
            )[0]
        else:
            params["constraint_value"] = random.choices(
                [
                    {"window": 10, "threshold": -0.1},
                    {"window": 28, "threshold": -0.1},
                    {"window": 364, "threshold": -0.01},
                    {"window": 10, "threshold": -0.1},
                    {"window": 10, "threshold": 0.01},
                ],
                [0.5, 0.2, 0.1, 0.2, 0.1],
            )[0]
    elif method_choice in ["stdev", "stdev_min"]:
        params["constraint_value"] = random.choices(
            [1.0, 0.5, 2.0, 3.0, 4.0], [0.2, 0.2, 0.1, 0.2, 0.1]
        )[0]
    elif method_choice in ["dampening"]:
        params["constraint_value"] = random.choices(
            [0.99, 0.9, 0.8, 0.999, 0.98, 0.9999, 0.95],
            [0.5, 0.2, 0.1, 0.2, 0.1, 0.2, 0.1],
        )[0]
        params['constraint_direction'] = "upper"
        params["constraint_regularization"] = 1.0
    elif method_choice in ["round"]:
        params["constraint_value"] = random.choices([0, 2], [0.4, 0.6])[0]
        # not used, just setting to fixed
        params['constraint_direction'] = "upper"
        params["constraint_regularization"] = 1.0
    elif method_choice in ["absolute", "fixed"]:
        params["constraint_value"] = random.choices([0, 0.1, 1], [0.8, 0.1, 0.1])[0]
    elif method_choice in ["historic_growth"]:
        params["constraint_value"] = {
            "threshold": random.choices([1.0, 0.5, 2.0, 0.2], [0.6, 0.2, 0.2, 0.04])[0]
        }
        window_choice = random.choices(
            [None, 10, 100, 360, 4], [0.8, 0.2, 0.2, 0.2, 0.04]
        )[0]
        if window_choice is not None:
            params["constraint_value"]["window"] = window_choice
        quantile_choice = random.choices(
            [None, 0.99, 0.9, 0.98, 0.75], [0.6, 0.2, 0.2, 0.2, 0.04]
        )[0]
        if quantile_choice is not None:
            params["constraint_value"]["quantile"] = quantile_choice
    elif method_choice in ["historic_diff"]:
        params["constraint_value"] = random.choices(
            [1.0, 0.5, 2.0, 0.2, 1.2],
            [0.6, 0.2, 0.2, 0.04, 0.04],
        )[0]
    return params


def fit_constraint(
    constraint_method,
    constraint_value,
    constraint_direction='upper',
    constraint_regularization=1.0,
    bounds=True,
    df_train=None,
    forecast_length=None,
):
    # check if training data provided
    if df_train is None and constraint_method in [
        "quantile",
        "stdev",
        "stdev_min",
        "last_window",
        "slope",
    ]:
        raise ValueError("this constraint requires df_train to be provided")
    # set direction
    lower_constraint = None
    upper_constraint = None
    if constraint_direction == "lower":
        lower_constraint = True
    elif constraint_direction == "upper":
        upper_constraint = True
    else:
        raise ValueError(f"constraint_direction: {constraint_direction} invalid")
    train_min = None
    train_max = None
    if constraint_method == "stdev_min":
        train_std = df_train.std(axis=0)
        if lower_constraint is not None:
            train_min = df_train.min(axis=0) - (constraint_value * train_std)
        if upper_constraint is not None:
            train_max = df_train.max(axis=0) + (constraint_value * train_std)
    elif constraint_method == "stdev":
        train_std = df_train.std(axis=0)
        train_mean = df_train.mean(axis=0)
        if lower_constraint is not None:
            train_min = train_mean - (constraint_value * train_std)
        if upper_constraint is not None:
            train_max = train_mean + (constraint_value * train_std)
    elif constraint_method in ["absolute", "fixed"]:
        train_min = constraint_value
        train_max = constraint_value
    elif constraint_method == "quantile":
        if lower_constraint is not None:
            train_min = df_train.quantile(constraint_value, axis=0)
        if upper_constraint is not None:
            train_max = df_train.quantile(constraint_value, axis=0)
    elif constraint_method == "last_window":
        if isinstance(constraint_value, dict):
            window = constraint_value.get("window", 3)
            window_agg = constraint_value.get("window_agg", "mean")
            if upper_constraint is not None:
                threshold = constraint_value.get("threshold", 0.05)
            else:
                threshold = constraint_value.get("threshold", -0.05)
        else:
            window = 1
            window_agg = "mean"
            threshold = constraint_value
        if window_agg == "mean":
            end_o_data = df_train.iloc[-window:].mean(axis=0)
        elif window_agg == "max":
            end_o_data = df_train.iloc[-window:].max(axis=0)
        elif window_agg == "min":
            end_o_data = df_train.iloc[-window:].min(axis=0)
        else:
            raise ValueError(f"constraint window_agg not recognized: {window_agg}")
        train_min = train_max = end_o_data + end_o_data * threshold
    elif constraint_method == "slope":
        if isinstance(constraint_value, dict):
            window = constraint_value.get("window", 1)
            window_agg = constraint_value.get("window_agg", "mean")
            slope = constraint_value.get("slope", 0.05)
            threshold = constraint_value.get("threshold", None)
        else:
            window = 1
            window_agg = "mean"
            slope = constraint_value
            threshold = None
        # slope is given as a final max growth, NOT compounding
        changes = constant_growth_rate(forecast_length, slope)
        if window_agg == "mean":
            end_o_data = df_train.iloc[-window:].mean(axis=0)
        elif window_agg == "max":
            end_o_data = df_train.iloc[-window:].max(axis=0)
        elif window_agg == "min":
            end_o_data = df_train.iloc[-window:].min(axis=0)
        else:
            raise ValueError(f"constraint window_agg not recognized: {window_agg}")
        # have the slope start above a threshold to allow more volatility
        if threshold is not None:
            end_o_data = end_o_data + end_o_data * threshold
        train_min = train_max = np.nan_to_num(
            end_o_data.to_numpy()
            + end_o_data.to_numpy()[np.newaxis, :] * changes[:, np.newaxis]
        )
    elif constraint_method == "historic_growth":
        if isinstance(constraint_value, dict):
            window = constraint_value.get("window", forecast_length)
            threshold = constraint_value.get("threshold", 1.0)
            quantile = constraint_value.get("quantile", None)
        else:
            window = forecast_length
            threshold = constraint_value
            quantile = None
        if window is None:
            window = forecast_length
        rolling_diff = df_train.diff(periods=window)
        # calculate the growth rates (slopes) by dividing the differences by the window size
        slopes = rolling_diff / window
        if quantile is not None:
            slopes_max = slopes.quantile(quantile).to_numpy()
            slopes_min = slopes.quantile(1 - quantile).to_numpy()
        else:
            slopes_max = slopes.max().to_numpy()
            slopes_min = slopes.min().to_numpy()
        # find the maximum growth rate and maximum decline rate for each series
        # and apply a log growth rate to that (to better allow for peaks like holidays)
        t = np.arange(forecast_length + 1).reshape(-1, 1)
        t2 = np.log1p(t) + 1
        train_max = (
            df_train.iloc[-1].to_numpy()
            + (((slopes_max * forecast_length)) * t2 / t2.max())[1:] * threshold
        )
        train_min = (
            df_train.iloc[-1].to_numpy()
            + (((slopes_min * forecast_length)) * t2 / t2.max())[1:] * threshold
        )
    elif constraint_method == "historic_diff":
        if isinstance(constraint_value, dict):
            threshold = constraint_value.get("threshold", 1.0)
        else:
            threshold = constraint_value
        rolling_diff = df_train.diff(periods=forecast_length)

        train_max = df_train.iloc[-1] + rolling_diff.max() * threshold
        train_min = df_train.iloc[-1] + rolling_diff.min() * threshold
    elif constraint_method == "dampening":
        pass
    elif constraint_method == "round":
        pass
    else:
        raise ValueError(
            f"constraint_method {constraint_method} not recognized, adjust constraint"
        )
    return lower_constraint, upper_constraint, train_min, train_max


def apply_fit_constraint(
    forecast,
    lower_forecast,
    upper_forecast,
    constraint_method,
    constraint_value,
    constraint_direction='upper',
    constraint_regularization=1.0,
    bounds=True,
    lower_constraint=None,
    upper_constraint=None,
    train_min=None,
    train_max=None,
    fillna=None,  # only used with regularization of 1 / None
):
    if constraint_method == "dampening":
        # the idea is to make the forecast plateau by gradually forcing the step to step change closer to zero
        trend_phi = constraint_value
        if trend_phi is not None and trend_phi != 1 and forecast.shape[0] > 2:
            req_len = forecast.shape[0] - 1
            phi_series = pd.Series(
                [trend_phi] * req_len,
                index=forecast.index[1:],
            ).pow(range(req_len))

            # adjust all by same margin
            forecast = pd.concat(
                [forecast.iloc[0:1], forecast.diff().iloc[1:].mul(phi_series, axis=0)]
            ).cumsum()

            if bounds:
                lower_forecast = pd.concat(
                    [
                        lower_forecast.iloc[0:1],
                        lower_forecast.diff().iloc[1:].mul(phi_series, axis=0),
                    ]
                ).cumsum()
                upper_forecast = pd.concat(
                    [
                        upper_forecast.iloc[0:1],
                        upper_forecast.diff().iloc[1:].mul(phi_series, axis=0),
                    ]
                ).cumsum()
        return forecast, lower_forecast, upper_forecast
    elif constraint_method == "round":
        decimals = int(constraint_value)
        if bounds:
            return (
                forecast.round(decimals=decimals),
                lower_forecast.round(decimals=decimals),
                upper_forecast.round(decimals=decimals),
            )
        else:
            return forecast.round(decimals=decimals), lower_forecast, upper_forecast
    if constraint_regularization == 1 or constraint_regularization is None:
        if fillna in [None, "None", "none", ""]:
            if lower_constraint is not None:
                forecast = forecast.clip(lower=train_min, axis=1)
            if upper_constraint is not None:
                forecast = forecast.clip(upper=train_max, axis=1)
            if bounds:
                if lower_constraint is not None:
                    lower_forecast = lower_forecast.clip(lower=train_min, axis=1)
                    upper_forecast = upper_forecast.clip(lower=train_min, axis=1)
                if upper_constraint is not None:
                    lower_forecast = lower_forecast.clip(upper=train_max, axis=1)
                    upper_forecast = upper_forecast.clip(upper=train_max, axis=1)
        else:
            # if FILLNA present, don't clip but replace with NaN then fill NaN
            if lower_constraint is not None:
                forecast = forecast.where(
                    forecast >= train_min,
                    np.nan,
                )
            if upper_constraint is not None:
                forecast = forecast.where(
                    forecast <= train_max,
                    np.nan,
                )
            if bounds:
                if lower_constraint is not None:
                    lower_forecast = lower_forecast.where(
                        lower_forecast >= train_min,
                        np.nan,
                    )
                    upper_forecast = upper_forecast.where(
                        upper_forecast >= train_min,
                        np.nan,
                    )
                if upper_constraint is not None:
                    lower_forecast = lower_forecast.where(
                        lower_forecast <= train_max,
                        np.nan,
                    )

                    upper_forecast = upper_forecast.where(
                        lower_forecast <= train_max, np.nan
                    )
            forecast = FillNA(forecast, method=str(fillna), window=10)
            if bounds:
                lower_forecast = FillNA(lower_forecast, method=str(fillna), window=10)
                upper_forecast = FillNA(upper_forecast, method=str(fillna), window=10)
    else:
        if lower_constraint is not None:
            forecast = forecast.where(
                forecast >= train_min,
                forecast + (train_min - forecast) * constraint_regularization,
            )
        if upper_constraint is not None:
            forecast = forecast.where(
                forecast <= train_max,
                forecast + (train_max - forecast) * constraint_regularization,
            )
        if bounds:
            if lower_constraint is not None:
                lower_forecast = lower_forecast.where(
                    lower_forecast >= train_min,
                    lower_forecast
                    + (train_min - lower_forecast) * constraint_regularization,
                )
                upper_forecast = upper_forecast.where(
                    upper_forecast >= train_min,
                    upper_forecast
                    + (train_min - upper_forecast) * constraint_regularization,
                )
            if upper_constraint is not None:
                lower_forecast = lower_forecast.where(
                    lower_forecast <= train_max,
                    lower_forecast
                    + (train_max - lower_forecast) * constraint_regularization,
                )

                upper_forecast = upper_forecast.where(
                    upper_forecast <= train_max,
                    upper_forecast
                    + (train_max - upper_forecast) * constraint_regularization,
                )
    return forecast, lower_forecast, upper_forecast


def apply_constraint_single(
    forecast,
    lower_forecast,
    upper_forecast,
    constraint_method,
    constraint_value,
    constraint_direction='upper',
    constraint_regularization=1.0,
    bounds=True,
    df_train=None,
):
    # note the Constraint Transformer also uses the same API so adjust changes there too
    lower_constraint, upper_constraint, train_min, train_max = fit_constraint(
        constraint_method=constraint_method,
        constraint_value=constraint_value,
        constraint_direction=constraint_direction,
        constraint_regularization=constraint_regularization,
        bounds=bounds,
        df_train=df_train,
        forecast_length=forecast.shape[0],
    )
    return apply_fit_constraint(
        forecast=forecast,
        lower_forecast=lower_forecast,
        upper_forecast=upper_forecast,
        constraint_method=constraint_method,
        constraint_value=constraint_value,
        constraint_direction=constraint_direction,
        constraint_regularization=constraint_regularization,
        bounds=bounds,
        lower_constraint=lower_constraint,
        upper_constraint=upper_constraint,
        train_min=train_min,
        train_max=train_max,
    )



# -*- coding: utf-8 -*-
"""Calendar conversion functions.

Includes Lunar, Chinese lunar, and Arabic lunar
"""
import numpy as np
import pandas as pd
from autots.tools.lunar import moon_phase_df


def lunar_from_lunar(new_moon):
    """Assumes continuous daily data and pre-needed start."""
    new_moon_dates = new_moon[new_moon == 1]
    # assuming the rule "first new moon on or after January 21st"
    filtered = new_moon_dates[
        ~((new_moon_dates.index.month == 1) & (new_moon_dates.index.day < 21))
    ].to_frame()
    filtered['year'] = filtered.index.year
    filtered['datetime'] = filtered.index
    new_years = filtered.groupby('year')['datetime'].min()
    new_years = pd.Series(1, index=new_years, name='new_years').to_frame()
    new_years['syear'] = np.arange(0, new_years.shape[0])
    new_years = pd.concat([new_moon_dates, new_years], axis=1)
    new_years['syear'] = new_years['syear'].ffill()
    new_years['lunar_month'] = new_years.groupby('syear').cumcount() + 1
    return new_years


def lunar_from_lunar_full(full_moon):
    """Assumes continuous daily data and pre-needed start."""
    full_moon_dates = full_moon[full_moon == 1]
    # assuming the rule "first full moon on or after January 21st"
    filtered = full_moon_dates[
        ~((full_moon_dates.index.month <= 3) & (full_moon_dates.index.day < 23))
    ].to_frame()
    filtered['year'] = filtered.index.year
    filtered['datetime'] = filtered.index
    new_years = filtered.groupby('year')['datetime'].min()
    new_years = pd.Series(1, index=new_years, name='new_years').to_frame()
    new_years['syear'] = np.arange(0, new_years.shape[0])
    new_years = pd.concat([full_moon_dates, new_years], axis=1)
    new_years['syear'] = new_years['syear'].ffill()
    new_years['lunar_month'] = new_years.groupby('syear').cumcount() + 1
    return new_years


def gregorian_to_christian_lunar(datetime_index):
    """Convert a pandas DatetimeIndex to Christian Lunar calendar. Aspiration it doesn't work exactly."""
    if isinstance(datetime_index, (str, list)):
        datetime_input = pd.to_datetime(datetime_index).sort_values()
    else:
        datetime_input = datetime_index.sort_values()
    expanded_dates = pd.date_range(
        datetime_input[0] - pd.Timedelta(days=365), datetime_input[-1], freq='D'
    )
    min_year = np.min(expanded_dates.year)
    moon_df = moon_phase_df(expanded_dates, epoch=2444238.5)
    lunar_months = lunar_from_lunar_full(moon_df['full_moon'])
    expanded_dates = pd.concat(
        [pd.Series(0, index=expanded_dates, name="date"), lunar_months], axis=1
    )
    expanded_dates['syear'] = expanded_dates['syear'].ffill()
    expanded_dates['lunar_month'] = expanded_dates['lunar_month'].ffill()
    expanded_dates['lunar_day'] = (
        expanded_dates.groupby(['syear', 'lunar_month']).cumcount() + 1
    )
    expanded_dates['lunar_year'] = expanded_dates['syear'] + min_year
    expanded_dates["weekofmonth"] = (expanded_dates["lunar_day"] - 1) // 7 + 1
    expanded_dates['dayofweek'] = expanded_dates.index.dayofweek
    return expanded_dates.loc[
        datetime_index,
        ['lunar_year', 'lunar_month', 'lunar_day', 'weekofmonth', 'dayofweek'],
    ].astype(int)


def gregorian_to_chinese(datetime_index):
    """Convert a pandas DatetimeIndex to Chinese Lunar calendar. Potentially has errors."""
    if isinstance(datetime_index, (str, list)):
        datetime_input = pd.to_datetime(datetime_index).sort_values()
    else:
        datetime_input = datetime_index.sort_values()
    expanded_dates = pd.date_range(
        datetime_input[0] - pd.Timedelta(days=365), datetime_input[-1], freq='D'
    )
    min_year = np.min(expanded_dates.year)
    moon_df = moon_phase_df(expanded_dates, epoch=2444238.5)
    lunar_months = lunar_from_lunar(moon_df['new_moon'])
    expanded_dates = pd.concat(
        [pd.Series(0, index=expanded_dates, name="date"), lunar_months], axis=1
    )
    expanded_dates['syear'] = expanded_dates['syear'].ffill()
    expanded_dates['lunar_month'] = expanded_dates['lunar_month'].ffill()
    expanded_dates['lunar_day'] = (
        expanded_dates.groupby(['syear', 'lunar_month']).cumcount() + 1
    )
    expanded_dates['lunar_year'] = expanded_dates['syear'] + min_year
    return (
        expanded_dates.loc[datetime_index, ['lunar_year', 'lunar_month', 'lunar_day']]
        .astype(int)
        .rename_axis(index='date')
    )


def to_jd(year, month, day):
    """Determine Julian day count from Islamic date. From convertdate by fitnr."""
    return (
        day
        + np.ceil(29.5 * (month - 1))
        + (year - 1) * 354
        + np.floor((3 + (11 * year)) / 30)
        + 1948439.5
    ) - 1


def gregorian_to_islamic(date, epoch_adjustment=1.5):
    """Calculate Islamic dates for pandas DatetimeIndex. Approximately. From convertdate by fitnr.

    Args:
        epoch_adjustment (float): 1.0 and that needs to be adjusted by about +/- 0.5 to account for timezone
    """
    if isinstance(date, (str, list)):
        date = pd.to_datetime(date)
    jd = date.to_julian_date()
    jd = np.floor(jd) + epoch_adjustment
    year = np.floor(((30 * (jd - 1948439.5)) + 10646) / 10631)
    month = np.minimum(12, np.ceil((jd - (29 + to_jd(year, 1, 1))) / 29.5) + 1)
    day = (jd - to_jd(year, month, 1)).astype(int) + 1
    return (
        pd.DataFrame({'year': year, 'month': month, 'day': day}, index=date)
        .astype(int)
        .rename_axis(index='date')
    )


def heb_is_leap(year):
    if (((7 * year) + 1) % 19) < 7:
        return True
    return False


def _elapsed_months(year):
    return (235 * year - 234) // 19


def _elapsed_days(year):
    months_elapsed = _elapsed_months(year)
    parts_elapsed = 204 + 793 * (months_elapsed % 1080)
    hours_elapsed = (
        5 + 12 * months_elapsed + 793 * (months_elapsed // 1080) + parts_elapsed // 1080
    )
    conjunction_day = 1 + 29 * months_elapsed + hours_elapsed // 24
    conjunction_parts = 1080 * (hours_elapsed % 24) + parts_elapsed % 1080

    if (
        (conjunction_parts >= 19440)
        or (
            (conjunction_day % 7 == 2)
            and (conjunction_parts >= 9924)
            and not heb_is_leap(year)
        )
        or (
            (conjunction_day % 7 == 1)
            and conjunction_parts >= 16789
            and heb_is_leap(year - 1)
        )
    ):
        alt_day = conjunction_day + 1
    else:
        alt_day = conjunction_day
    if alt_day % 7 in [0, 3, 5]:
        alt_day += 1

    return alt_day


def _days_in_year(year):
    return _elapsed_days(year + 1) - _elapsed_days(year)


def _long_cheshvan(year):
    """Returns True if Cheshvan has 30 days"""
    return _days_in_year(year) % 10 == 5


def _short_kislev(year):
    """Returns True if Kislev has 29 days"""
    return _days_in_year(year) % 10 == 3


def _month_length(year, month):
    """Months start with Nissan (Nissan is 1 and Tishrei is 7)"""
    if month in [1, 3, 5, 7, 11]:
        return 30
    if month in [2, 4, 6, 10, 13]:
        return 29
    if month == 12:
        if heb_is_leap(year):
            return 30
        return 29
    if month == 8:  # if long Cheshvan return 30, else return 29
        if _long_cheshvan(year):
            return 30
        return 29
    if month == 9:  # if short Kislev return 29, else return 30
        if _short_kislev(year):
            return 29
        return 30
    raise ValueError("Invalid month")


def gregorian_to_hebrew(dates):
    """Convert pd.Datetimes to a Hebrew date. From pyluach by simlist.

    This is the slowest of the lot and needs to be improved.
    """
    if isinstance(dates, (str, list)):
        day = pd.to_datetime(dates).to_julian_date()
    else:
        day = dates.to_julian_date()
    if (day <= 347997).any():
        raise ValueError("According to this calendar, this time doesn't exist")

    jd = (day + 0.5).astype(int)  # Try to account for half day
    jd -= 347997
    years = (jd // 365).astype(int) + 2  # try that to debug early years
    date_list = []
    for idx in range(len(jd)):
        c_jd = jd[idx]
        year = years[idx]
        first_day = _elapsed_days(year)

        while first_day > c_jd:
            year -= 1
            first_day = _elapsed_days(year)

        months = [7, 8, 9, 10, 11, 12, 13, 1, 2, 3, 4, 5, 6]
        if not heb_is_leap(year):
            months.remove(13)

        days_remaining = c_jd - first_day
        for month in months:
            if days_remaining >= _month_length(year, month):
                days_remaining -= _month_length(year, month)
            else:
                date_list.append(
                    pd.DataFrame(
                        {"year": year, "month": month, 'day': days_remaining + 1},
                        index=[dates[idx]],
                    )
                )
                break
    return pd.concat(date_list, axis=0).rename_axis(index='date')


def gregorian_to_hindu(datetime_index):
    """Convert a pandas DatetimeIndex to Hindu calendar date components.
    Hindu calendar has numerous regional variations.

    Used an llm to put this one together.
    It gets the dates wrong, but it does appear to have correlated consistency so may still work for modeling.
    Suggestions for improvement welcomed.
    """
    if isinstance(datetime_index, (str, list)):
        datetime_input = pd.to_datetime(datetime_index).sort_values()
    else:
        datetime_input = datetime_index.sort_values()
    # Expand date range to cover previous year for new moons
    expanded_dates = pd.date_range(
        datetime_input[0] - pd.Timedelta(days=365), datetime_input[-1], freq='D'
    )
    min_year = np.min(expanded_dates.year)
    # Get moon phases
    moon_df = moon_phase_df(expanded_dates, epoch=2444238.5)
    # Use new moon dates to define lunar months (Amanta system)
    lunar_months = lunar_from_lunar(moon_df['new_moon'])
    # Merge with expanded dates
    expanded_dates = pd.concat(
        [pd.Series(0, index=expanded_dates, name="date"), lunar_months], axis=1
    )
    expanded_dates['syear'] = expanded_dates['syear'].ffill()
    expanded_dates['lunar_month'] = expanded_dates['lunar_month'].ffill()
    # Calculate lunar day (tithi)
    expanded_dates['lunar_day'] = (
        expanded_dates.groupby(['syear', 'lunar_month']).cumcount() + 1
    )
    expanded_dates['lunar_year'] = expanded_dates['syear'] + min_year
    # Assign approximate Hindu month names
    hindu_month_names = {
        1: 'Chaitra',
        2: 'Vaishakha',
        3: 'Jyeshtha',
        4: 'Ashadha',
        5: 'Shravana',
        6: 'Bhadrapada',
        7: 'Ashwin',
        8: 'Kartika',
        9: 'Margashirsha',
        10: 'Pausha',
        11: 'Magha',
        12: 'Phalguna',
    }
    # Adjust lunar_month to fit within 12 months
    expanded_dates['hindu_month_number'] = (
        (expanded_dates['lunar_month'] - 1) % 12
    ) + 1
    expanded_dates['hindu_month_name'] = expanded_dates['hindu_month_number'].map(
        hindu_month_names
    )
    # Return the data for the input dates
    return expanded_dates.loc[
        datetime_input,
        ['lunar_year', 'hindu_month_number', 'hindu_month_name', 'lunar_day'],
    ].rename_axis(index='date')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  5 08:21:47 2025

@author: colincatlin
"""

import unittest
import numpy as np
import pandas as pd

from autots.evaluator.validation import (
    extract_seasonal_val_periods,
    validate_num_validations,
    generate_validation_indices,
)

# Create a dummy DataFrame to use in tests.
def create_dummy_df(n_rows=10, n_cols=1):
    # Using a simple integer index here for clarity.
    data = np.random.randn(n_rows, n_cols)
    return pd.DataFrame(data, index=pd.date_range("2020-01-01", freq='D', periods=n_rows))


class TestValidationSegments(unittest.TestCase):

    def test_extract_seasonal_val_periods(self):
        # A string with digits should return the integer value
        self.assertEqual(extract_seasonal_val_periods("seasonal 364"), 364)
        self.assertEqual(extract_seasonal_val_periods("seasonal6"), 6)
        self.assertEqual(extract_seasonal_val_periods("s4eas3onal"), 43)
        # If no digits are found, int('') will raise a ValueError.
        with self.assertRaises(ValueError):
            extract_seasonal_val_periods("seasonal")  # no digits

    def test_validate_num_validations_backwards(self):
        # Using a non‐seasonal method, e.g. "backwards"
        df = create_dummy_df(n_rows=10)
        forecast_length = 2

        # For non‐seasonal methods, max_possible = df_rows / forecast_length.
        # Here 10/2 = 5.0; since 5.0 - 5 = 0 (not > min_allowed_train_percent),
        # max_possible becomes int(5) - 1 = 4.
        # When num_validations=2: since 4 >= (2+1), the input is accepted.
        result = validate_num_validations(
            validation_method="backwards",
            num_validations=2,
            df_wide_numeric=df,
            forecast_length=forecast_length,
            min_allowed_train_percent=0.5,
            verbose=0,
        )
        self.assertEqual(result, 2)

        # If too many validations are requested, it is reduced.
        result = validate_num_validations(
            validation_method="backwards",
            num_validations=5,
            df_wide_numeric=df,
            forecast_length=forecast_length,
            min_allowed_train_percent=0.5,
            verbose=0,
        )
        # Here 4 < (5+1)=6, so the function should reduce num_validations to max_possible - 1 = 4 - 1 = 3.
        self.assertEqual(result, 3)

    def test_validate_num_validations_seasonal(self):
        # For seasonal validations with extra period specified.
        df = create_dummy_df(n_rows=10)
        forecast_length = 2
        # For validation_method like "seasonal 4": the seasonal period is 4.
        # temp = df_rows + forecast_length = 10+2 = 12, so max_possible = 12/4 = 3.0.
        # With no fractional part > min_allowed_train_percent, max_possible becomes 3 - 1 = 2.
        # Now if we ask for 2 validations then 2 < (2+1)=3 so it is reduced to 2-1 = 1.
        result = validate_num_validations(
            validation_method="seasonal 4",
            num_validations=2,
            df_wide_numeric=df,
            forecast_length=forecast_length,
            min_allowed_train_percent=0.5,
            verbose=0,
        )
        self.assertEqual(result, 1)

        # When using "auto" or "max", check the returned values.
        result_auto = validate_num_validations(
            validation_method="seasonal 4",
            num_validations="auto",
            df_wide_numeric=df,
            forecast_length=forecast_length,
            min_allowed_train_percent=0.5,
            verbose=0,
        )
        # With max_possible = 2, the auto branch returns max_possible (i.e. 2).
        self.assertEqual(result_auto, 2)

        result_max = validate_num_validations(
            validation_method="seasonal 4",
            num_validations="max",
            df_wide_numeric=df,
            forecast_length=forecast_length,
            min_allowed_train_percent=0.5,
            verbose=0,
        )
        # In the "max" branch, returns max_possible - 1 = 2 - 1 = 1.
        self.assertEqual(result_max, 1)

    def test_generate_validation_indices_similarity(self):
        # For similarity method, we patch GeneralTransformer and retrieve_closest_indices.
        df = create_dummy_df(n_rows=100)
        forecast_length = 1
        num_validations = 1
        validation_method = "similarity"
        validation_params = {}

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            validation_params=validation_params,
            verbose=0,
        )

        # We expect a list of length 2.
        self.assertEqual(len(result), 2)

    def test_generate_validation_indices_seasonal(self):
        # it might be worth rewriting this without the @patch

        # For seasonal method without extra period (i.e. exactly "seasonal")
        df = create_dummy_df(n_rows=100)
        forecast_length = 2
        num_validations = 2
        validation_method = "seasonal"
        validation_params = {}

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            validation_params=validation_params,
            verbose=0,
        )

        self.assertEqual(len(result), 3)
        # This doesn't test much of anything except that the default is the same, if default changes, change this
        self.assertEqual(len(list(result[0])), 94)

    def test_generate_validation_indices_backwards(self):
        # Test the backwards method (e.g., "backwards" or its variants)
        df = create_dummy_df(n_rows=10)
        forecast_length = 2
        num_validations = 2
        validation_method = "backwards"

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            verbose=0,
        )
        # For "backwards", the code first sets validation_indexes = [df.index]
        # and then for each validation appends idx[0: n_rows - (y+1)*forecast_length].
        # With n_rows=10 and forecast_length=2:
        #  y=0 -> slice = df.index[0:10-2] = df.index[0:8]
        #  y=1 -> slice = df.index[0:10-4] = df.index[0:6]
        self.assertEqual(len(result), 3)
        self.assertEqual(list(result[0]), list(df.index))
        self.assertEqual(list(result[1]), list(df.index[0:8]))
        self.assertEqual(list(result[2]), list(df.index[0:6]))

    def test_generate_validation_indices_even(self):
        # Test the even method
        df = create_dummy_df(n_rows=10)
        forecast_length = 2
        num_validations = 2
        validation_method = "even"

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            verbose=0,
        )
        # For "even", the initial index is [df.index],
        # and then for y in range(num_validations):
        #   validation_size = floor((n_rows - forecast_length)/(num_validations+1))
        # Here: (10-2)/(3) = 8/3 -> floor(2.66) = 2.
        # Then:
        #   y=0: slice = df.index[0: 2*1 + 2] = df.index[0:4]
        #   y=1: slice = df.index[0: 2*2 + 2] = df.index[0:6]
        self.assertEqual(len(result), 3)
        self.assertEqual(list(result[0]), list(df.index))
        self.assertEqual(list(result[1]), list(df.index[0:4]))
        self.assertEqual(list(result[2]), list(df.index[0:6]))

    def test_generate_validation_indices_seasonal_with_period(self):
        # Test the seasonal method when extra digits are provided (e.g., "seasonal 4")
        df = create_dummy_df(n_rows=10)
        forecast_length = 2
        num_validations = 2
        validation_method = "seasonal 4"

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            verbose=0,
        )
        # Because "seasonal 4" is caught by the base-val branch,
        # the initial validation_indexes = [df.index]
        # then in the seasonal branch:
        #   seasonal_val_periods = 4.
        # For y=0: val_per = (1*4) - forecast_length = 4 - 2 = 2; then slice index = df.index[0: (10-2)] = df.index[0:8]
        # For y=1: val_per = (2*4) - forecast_length = 8 - 2 = 6; then slice index = df.index[0: (10-6)] = df.index[0:4]
        expected = [
            df.index,
            df.index[0:8],
            df.index[0:4],
        ]
        self.assertEqual(len(result), 3)
        for res, exp in zip(result, expected):
            self.assertEqual(list(res), list(exp))

    def test_generate_validation_indices_mixed_length(self):
        # Test the mixed_length validation method.
        df = create_dummy_df(n_rows=10)
        forecast_length = 2
        num_validations = 2
        validation_method = "mixed_length"

        result = generate_validation_indices(
            validation_method=validation_method,
            forecast_length=forecast_length,
            num_validations=num_validations,
            df_wide_numeric=df,
            verbose=0,
        )
        # For mixed_length, the loop iterates num_validations+1 times (i.e. 3 times):
        # count == 0: cut = int(10/2)=5, so tuple: (df.index[0:5], df.index[5:])
        # count == 1: cut = 10 - int(10/3)= 10 - 3 =7, so tuple: (df.index[0:7], df.index[7:])
        # count == 2: cut = 10 - (3*2)= 10 - 6 =4,
        #          so tuple: (df.index[0:4], df.index[4: 4+2] = df.index[4:6])
        self.assertEqual(len(result), 3)
        expected = [
            (df.index[0:5], df.index[5:]),
            (df.index[0:7], df.index[7:]),
            (df.index[0:4], df.index[4:6]),
        ]
        for (res1, res2), (exp1, exp2) in zip(result, expected):
            self.assertEqual(list(res1), list(exp1))
            self.assertEqual(list(res2), list(exp2))

    def test_generate_validation_indices_unknown(self):
        # If an unknown validation_method is provided, the function should raise a ValueError.
        df = create_dummy_df(n_rows=100)
        forecast_length = 2
        num_validations = 2
        validation_method = "unknown_method"

        with self.assertRaises(ValueError):
            generate_validation_indices(
                validation_method=validation_method,
                forecast_length=forecast_length,
                num_validations=num_validations,
                df_wide_numeric=df,
                verbose=0,
            )

    def test_validate_num_validations_short_data(self):
        # Test that if the data is too short to allow any validations, 0 is returned.
        df = create_dummy_df(n_rows=1)
        forecast_length = 2  # longer than data
        result = validate_num_validations(
            validation_method="backwards",
            num_validations=2,
            df_wide_numeric=df,
            forecast_length=forecast_length,
            verbose=0,
        )
        self.assertEqual(result, 0)

    def test_validate_num_validations_negative_input(self):
        # If a negative number is provided, abs() is taken and the result is adjusted if needed.
        df = create_dummy_df(n_rows=100)
        forecast_length = 2
        # Passing -1 will be converted to 1.
        result = validate_num_validations(
            validation_method="backwards",
            num_validations=-1,
            df_wide_numeric=df,
            forecast_length=forecast_length,
            verbose=0,
        )
        self.assertEqual(result, 1)


# if __name__ == "__main__":
#     unittest.main()



"""
Complicated
"""

import random
import datetime
import numpy as np
import pandas as pd
from autots.models.base import ModelObject, PredictionObject
from autots.tools.probabilistic import Point_to_Probability
from autots.tools.percentile import nan_quantile, trimmed_mean
from autots.tools.transform import (
    GeneralTransformer,
    RandomTransform,
)
from autots.models.model_list import scalable


class PreprocessingExperts(ModelObject):
    """Regression use the last n values as the basis of training data.

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
        # regression_type: str = None,
    """

    def __init__(
        self,
        name: str = "PreprocessingExperts",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        holiday_country: str = 'US',
        random_seed: int = 2023,
        verbose: int = 0,
        model_params: dict = {
            "model_str": 'DatepartRegresion',
            "parameter_dict": {
                'regression_model': {
                    'model': 'ElasticNet',
                    'model_params': {
                        'l1_ratio': 0.1,
                        'fit_intercept': True,
                        'selection': 'cyclic',
                        'max_iter': 1000,
                    },
                },
                'datepart_method': 'expanded_binarized',
                'polynomial_degree': None,
                'holiday_countries_used': False,
                'lags': None,
                'forward_lags': None,
                'regression_type': None,
            },
            "transformation_dict": {
                'fillna': 'rolling_mean_24',
                'transformations': {0: 'StandardScaler'},
                'transformation_params': {0: {}},
            },
        },
        transformation_dict=None,
        forecast_length: int = 28,
        point_method: str = "mean",
        n_jobs: int = -1,
        **kwargs,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            holiday_country=holiday_country,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
        )
        self.model_params = model_params
        self.transformation_dict = transformation_dict
        self.point_method = point_method
        self.forecast_length = forecast_length

    def fit(self, df, future_regressor=None):
        """Train algorithm given data supplied.

        Args:
            df (pandas.DataFrame): Datetime Indexed
        """
        self.fit_data(df)
        # from autots.tools.transform import GeneralTransformer  # avoid circular imports

        if self.transformation_dict is None:
            raise ValueError(
                "transformation_dict cannot be None with PreprocessingRegression"
            )

        self.transformer_object = GeneralTransformer(
            n_jobs=self.n_jobs,
            holiday_country=self.holiday_country,
            verbose=self.verbose,
            random_seed=self.random_seed,
            forecast_length=self.forecast_length,
            **self.transformation_dict,
        )
        df = self.transformer_object._first_fit(df)

        from autots.evaluator.auto_model import (
            ModelPrediction,
        )  # avoid circular imports

        new_df = df
        self.model = {}
        trans_keys = sorted(list(self.transformer_object.transformations.keys()))

        self.model[0] = ModelPrediction(
            forecast_length=self.forecast_length,
            frequency=self.frequency,
            **self.model_params,
        ).fit(new_df, future_regressor=future_regressor)
        for i in trans_keys:
            trans_name = self.transformer_object.transformations[i]
            # assumes first i is always 0
            if trans_name not in ['Slice']:
                new_df = self.transformer_object._fit_one(new_df, i)
            self.model[int(i) + 1] = ModelPrediction(
                forecast_length=self.forecast_length,
                frequency=self.frequency,
                **self.model_params,
            ).fit(new_df, future_regressor=future_regressor)

        self.fit_runtime = datetime.datetime.now() - self.startTime
        return self

    def fit_data(self, df, future_regressor=None):
        df = self.basic_profile(df)
        return self

    def predict(
        self,
        forecast_length: int = None,
        future_regressor=None,
        just_point_forecast: bool = False,
        df=None,
    ):
        """Generate forecast data immediately following dates of .fit().

        Args:
            forecast_length (int): Number of periods of data to forecast ahead
            regressor (numpy.Array): additional regressor, not used
            just_point_forecast (bool): If True, return a pandas.DataFrame of just point forecasts

        Returns:
            Either a PredictionObject of forecasts and metadata, or
            if just_point_forecast == True, a dataframe of point forecasts
        """
        predictStartTime = datetime.datetime.now()
        if df is not None:
            self.fit_data(df)
        if forecast_length is None:
            forecast_length = self.forecast_length
        if int(forecast_length) > int(self.forecast_length) and not self.one_step:
            print("Regression must be refit to change forecast length!")
        index = self.create_forecast_index(forecast_length=forecast_length)

        df_list = []
        self.saved = {}
        trans_keys = sorted(list(self.transformer_object.transformations.keys()))
        # first one is on no preprocessing
        rfPred = self.model[0].predict(
            forecast_length, future_regressor=future_regressor
        )
        self.saved[0] = rfPred.forecast.copy()
        df_list.append(rfPred.forecast)
        for i in trans_keys:
            key = int(i) + 1
            rfPred = self.model[key].predict(
                forecast_length, future_regressor=future_regressor
            )
            self.saved[key] = rfPred.forecast.copy()
            # rfPred_upper = self.transformer_object.inverse_transform(rfPred.upper_forecast, start=i)
            # rfPred_lower = self.transformer_object.inverse_transform(rfPred.lower_forecast, start=i)
            rfPred = self.transformer_object.inverse_transform(rfPred.forecast, start=i)
            df_list.append(rfPred)
        # (num_windows, forecast_length, num_series)
        self.result_windows = np.asarray(df_list)

        if self.point_method == "weighted_mean":
            # weighted by later (more preprocessing higher)
            weights = np.arange(len(trans_keys) + 1) + 1
            forecast = np.average(self.result_windows, axis=0, weights=weights)
        elif self.point_method == "mean":
            forecast = np.nanmean(self.result_windows, axis=0)
        elif self.point_method == "median":
            forecast = np.nanmedian(self.result_windows, axis=0)
        elif self.point_method == "midhinge":
            q1 = nan_quantile(self.result_windows, q=0.25, axis=0)
            q2 = nan_quantile(self.result_windows, q=0.75, axis=0)
            forecast = (q1 + q2) / 2
        elif self.point_method == "trimmed_mean_20":
            forecast = trimmed_mean(self.result_windows, percent=0.2, axis=0)
        elif self.point_method == "trimmed_mean_40":
            forecast = trimmed_mean(self.result_windows, percent=0.4, axis=0)
        elif self.point_method == "closest":
            forecast = self.result_windows[-1]
        else:
            raise ValueError(f"point_method {self.point_method} not recognized.")

        df = pd.DataFrame(forecast, index=index, columns=self.column_names)
        pred_int = round((1 - self.prediction_interval) / 2, 5)
        upper_forecast = nan_quantile(self.result_windows, q=(1 - pred_int), axis=0)
        lower_forecast = nan_quantile(self.result_windows, q=pred_int, axis=0)
        lower_forecast = pd.DataFrame(
            lower_forecast, index=index, columns=self.column_names
        )
        upper_forecast = pd.DataFrame(
            upper_forecast, index=index, columns=self.column_names
        )
        if just_point_forecast:
            return df
        else:
            if False:
                # might still be worth using, as the spread of forecasts above is often small
                upper_forecast, lower_forecast = Point_to_Probability(
                    self.last_window,
                    df,
                    prediction_interval=self.prediction_interval,
                    method='inferred_normal',
                )

            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=forecast_length,
                forecast_index=df.index,
                forecast_columns=df.columns,
                lower_forecast=lower_forecast,
                forecast=df,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )
            return prediction

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        from autots.evaluator.auto_model import ModelMonster  # avoid circular imports

        model_type = random.choices(list(scalable.keys()), list(scalable.values()))[0]
        model_params = ModelMonster(model_type).get_new_params(method="default")
        return {
            "point_method": random.choices(
                [
                    "weighted_mean",
                    "mean",
                    "median",
                    "midhinge",
                    'closest',
                    'trimmed_mean_20',
                    'trimmed_mean_40',
                ],
                [0.2, 0.2, 0.2, 0.2, 0.1, 0.1, 0.1],
            )[0],
            'model_params': {
                "model_str": model_type,
                "parameter_dict": model_params,
                "transformation_dict": RandomTransform(
                    transformer_list="scalable",
                    transformer_max_depth=2,
                    allow_none=True,
                ),
            },
            'transformation_dict': None,  # assume this passed via AutoTS transformer dict
        }

    def get_params(self):
        """Return dict of current parameters."""
        return {
            "point_method": self.point_method,
            'model_params': self.model_params,
            'transformation_dict': self.transformation_dict,
        }



"""
Nixtla's NeuralForecast. Be warned, as of writing, their package has commercial restrictions.
"""

import logging
import random
import datetime
import pandas as pd
from autots.models.base import ModelObject, PredictionObject
from autots.tools.probabilistic import Point_to_Probability


class NeuralForecast(ModelObject):
    """See NeuralForecast documentation for details.


    temp['ModelParameters'].str.extract('model": "([a-zA-Z]+)')

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
        regression_type: str = None,
        model (str or object): string aliases or passed to the models arg of NeuralForecast
        model_args (dict): for all model args that aren't in default list, run get_new_params for default
    """

    def __init__(
        self,
        name: str = "NeuralForecast",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        holiday_country: str = 'US',
        random_seed: int = 2023,
        verbose: int = 0,
        forecast_length: int = 28,
        regression_type: str = None,
        n_jobs: int = 1,
        model="LSTM",
        loss="MQLoss",
        input_size="2ForecastLength",
        max_steps=100,
        learning_rate=0.001,
        early_stop_patience_steps=-1,
        activation='ReLU',
        scaler_type='robust',
        model_args={},
        point_quantile=None,
        **kwargs,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            holiday_country=holiday_country,
            random_seed=random_seed,
            regression_type=regression_type,
            verbose=verbose,
            n_jobs=n_jobs,
        )
        self.model = model
        self.loss = loss
        self.input_size = input_size
        self.max_steps = max_steps
        self.learning_rate = learning_rate
        self.early_stop_patience_steps = early_stop_patience_steps
        self.activation = activation
        self.scaler_type = scaler_type
        self.model_args = model_args
        self.point_quantile = point_quantile
        self.forecast_length = forecast_length
        self.df_train = None
        self.static_regressor = None

    def fit(
        self,
        df,
        future_regressor=None,
        static_regressor=None,
        regressor_per_series=None,
    ):
        """Train algorithm given data supplied.

        Args:
            df (pandas.DataFrame): Datetime Indexed
        """
        self.basic_profile(df)
        static_cols = None
        if self.regression_type in ["User", "user", True]:
            if future_regressor is None:
                raise ValueError(
                    "regression_type='User' but no future_regressor passed"
                )
            self.static_regressor = static_regressor
            if isinstance(self.static_regressor, pd.DataFrame):
                static_cols = static_regressor.columns.tolist()
            if regressor_per_series is not None:
                if not isinstance(regressor_per_series, dict):
                    raise ValueError("regressor_per_series in incorrect format")

        from neuralforecast import NeuralForecast
        from neuralforecast.losses.pytorch import (
            SMAPE,
            MQLoss,
            DistributionLoss,
            MAE,
            HuberLoss,
        )
        from neuralforecast.models import (
            TFT,
            LSTM,
            NHITS,
            MLP,
            NBEATS,
            TimesNet,
            PatchTST,
            DeepAR,
            TiDE,
            FEDformer,
        )

        prediction_interval = self.prediction_interval
        forecast_length = self.forecast_length
        freq = self.frequency

        # Split data and declare panel dataset
        if isinstance(prediction_interval, list):
            levels = prediction_interval
        elif isinstance(prediction_interval, dict):
            levels = list(prediction_interval.values())
        else:
            levels = [int(prediction_interval * 100)]

        logging.getLogger("pytorch_lightning").setLevel(logging.CRITICAL)
        loss = self.loss
        if loss == "MQLoss":
            if self.point_quantile is None:
                loss = MQLoss(level=levels)
            else:
                div = (1 - self.prediction_interval) / 2
                quantiles = [div, 1 - div, self.point_quantile]
                loss = MQLoss(quantiles=quantiles)
        elif loss == "HuberMQLoss":
            from neuralforecast.losses.pytorch import HuberMQLoss

            if self.point_quantile is None:
                loss = HuberMQLoss(level=levels)
            else:
                div = (1 - self.prediction_interval) / 2
                quantiles = [div, 1 - div, self.point_quantile]
                loss = HuberMQLoss(quantiles=quantiles)
        elif loss == "Poisson":
            loss = DistributionLoss(
                distribution='Poisson', level=levels, return_params=False
            )
        elif loss == "Bernoulli":
            loss = DistributionLoss(
                distribution='Bernoulli', level=levels, return_params=False
            )
        elif loss == "StudentT":
            loss = DistributionLoss(
                distribution='StudentT', level=levels, return_params=False
            )
        elif loss == "NegativeBinomial":
            loss = DistributionLoss(
                distribution='NegativeBinomial',
                level=levels,
                total_count=3,
                return_params=False,
            )
        elif loss == "Normal":
            loss = DistributionLoss(
                distribution='Normal', level=levels, return_params=False
            )
        elif loss == "Tweedie":
            loss = DistributionLoss(
                distribution='Tweedie', level=levels, return_params=False, rho=1.5
            )
        elif loss == "MAE":
            self.df_train = df
            loss = MAE()
        elif loss == "HuberLoss":
            self.df_train = df
            loss = HuberLoss()
        elif loss == "SMAPE":
            self.df_train = df
            loss = SMAPE()
        elif isinstance(loss, str):
            raise ValueError(f"loss not recognized: {loss}")
        else:
            # allow custom input
            pass

        str_input = str(self.input_size).lower()
        if "forecastlength" in str_input:
            input_size = forecast_length * int(
                ''.join([x for x in str_input if x.isdigit()])
            )
        else:
            input_size = int(self.input_size)
        self.base_args = {
            "h": forecast_length,
            "input_size": input_size,
            "max_steps": self.max_steps,
            # "num_workers_loader": self.n_jobs,
            "random_seed": self.random_seed,
            "learning_rate": self.learning_rate,
            "loss": loss,
            'scaler_type': self.scaler_type,
            # trying to suppress the stupid file logging lightning does
            "logger": False,
            "log_every_n_steps": 0,
        }
        models = self.model
        model_args = self.model_args
        if self.regression_type in ['User', 'user', True]:
            regr_cols = future_regressor.columns.tolist()
            if regressor_per_series is not None:
                regr_cols + next(iter(regressor_per_series.values())).columns.tolist()
            self.base_args["futr_exog_list"] = regr_cols
            self.base_args['stat_exog_list'] = static_cols

        if isinstance(models, list):
            # User inputs classes directly
            pass
        elif models == 'LSTM':
            models = [LSTM(**{**self.base_args, **model_args})]
        elif models == "NHITS":
            models = [NHITS(**{**self.base_args, **model_args})]
        elif models == "NBEATS":
            models = [NBEATS(**{**self.base_args, **model_args})]
        elif models == "MLP":
            models = [MLP(**{**self.base_args, **model_args})]
        elif models == "TimesNet":
            models = [TimesNet(**{**self.base_args, **model_args})]
        elif models == "TFT":
            models = [TFT(**{**self.base_args, **model_args})]
        elif models == "PatchTST":
            models = [PatchTST(**{**self.base_args, **model_args})]
        elif models == "DeepAR":
            models = [DeepAR(**{**self.base_args, **model_args})]
        elif models == "TiDE":
            models = [TiDE(**{**self.base_args, **model_args})]
        elif models == "FEDformer":
            models = [FEDformer(**{**self.base_args, **model_args})]
        else:
            raise ValueError(f"models not recognized: {models}")

        # model params
        # requires pandas >= 1.5
        silly_format = df.reset_index(names='ds').melt(
            id_vars='ds', value_name='y', var_name='unique_id'
        )
        if self.regression_type in ['User', 'user', True]:
            silly_format = silly_format.merge(
                future_regressor, left_on='ds', right_index=True
            )
            if regressor_per_series is not None:
                full_df = []
                for key, value in regressor_per_series.items():
                    local_copy = value.copy().reindex(df.index)
                    local_copy.index.name = 'ds'
                    local_copy = local_copy.reset_index()
                    local_copy['unique_id'] = str(key)
                    full_df.append(local_copy)
                silly_format = silly_format.merge(
                    pd.concat(full_df), on=['unique_id', 'ds'], how='left'
                ).fillna(0)
        self.nf = NeuralForecast(models=models, freq=freq)
        if self.static_regressor is None:
            self.nf.fit(df=silly_format)
        else:
            another_silly_format = self.static_regressor
            another_silly_format.index.name = "unique_id"
            self.nf.fit(
                df=silly_format, static_df=another_silly_format.reset_index(drop=False)
            )
        self.fit_runtime = datetime.datetime.now() - self.startTime
        return self

    def predict(
        self,
        forecast_length=None,
        future_regressor=None,
        just_point_forecast=False,
        regressor_per_series=None,
    ):
        predictStartTime = datetime.datetime.now()
        if self.regression_type in ['User', 'user', True]:
            index = self.create_forecast_index(forecast_length=self.forecast_length)
            futr_df = pd.concat(
                [
                    pd.Series(col, index=index, name='unique_id')
                    for col in self.column_names
                ]
            )
            futr_df = futr_df.to_frame().merge(
                future_regressor, left_index=True, right_index=True
            )
            futr_df = futr_df.reset_index(names='ds')
            if regressor_per_series is not None:
                full_df = []
                for key, value in regressor_per_series.items():
                    local_copy = value.copy().reindex(index)
                    local_copy.index.name = 'ds'
                    local_copy = local_copy.reset_index()
                    local_copy['unique_id'] = str(key)
                    full_df.append(local_copy)
                futr_df = futr_df.merge(
                    pd.concat(full_df), on=['unique_id', 'ds'], how='left'
                ).fillna(0)
            self.futr_df = futr_df
            long_forecast = self.nf.predict(futr_df=futr_df)
        else:
            long_forecast = self.nf.predict()
        # self.long_forecast = long_forecast
        target_col = [x for x in long_forecast.columns.tolist() if "median" in str(x)]
        if len(target_col) < 1:
            target_col = long_forecast.columns[-1]
        else:
            target_col = target_col[0]
        if self.point_quantile is not None:
            # print(long_forecast.columns)
            target_col = long_forecast.columns[-1]
        forecast = long_forecast.reset_index().pivot_table(
            index='ds', columns='unique_id', values=target_col
        )[self.column_names]

        if just_point_forecast:
            return forecast
        if self.df_train is not None:
            upper_forecast, lower_forecast = Point_to_Probability(
                self.df_train,
                forecast,
                method='inferred_normal',
                prediction_interval=self.prediction_interval,
            )
        else:
            target_col = [x for x in long_forecast.columns if "hi-" in x][0]
            # print(f"upper target col: {target_col}")
            upper_forecast = long_forecast.reset_index().pivot_table(
                index='ds', columns='unique_id', values=target_col
            )[self.column_names]
            target_col = [x for x in long_forecast.columns if "lo-" in x][0]
            # print(f"lower target col {target_col}")
            lower_forecast = long_forecast.reset_index().pivot_table(
                index='ds', columns='unique_id', values=target_col
            )[self.column_names]

        predict_runtime = datetime.datetime.now() - predictStartTime
        prediction = PredictionObject(
            model_name=self.name,
            forecast_length=self.forecast_length,
            forecast_index=forecast.index,
            forecast_columns=forecast.columns,
            lower_forecast=lower_forecast,
            forecast=forecast,
            upper_forecast=upper_forecast,
            prediction_interval=self.prediction_interval,
            predict_runtime=predict_runtime,
            fit_runtime=self.fit_runtime,
            model_parameters=self.get_params(),
        )
        return prediction

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        model_list = [
            'DeepAR',
            'MLP',
            "LSTM",
            "PatchTST",
            "NHITS",
            "TFT",
            "TimesNet",
            "TiDE",
            "FEDformer",
        ]
        if method in model_list:
            models = method
        else:
            models = random.choices(
                model_list, [0.05, 0.4, 0.2, 0.2, 0.2, 0.1, 0.1, 0.1, 0.1]
            )[0]
        if "regressor" in method:
            regression_type_choice = "User"
        else:
            regression_type_choice = random.choices([None, "User"], weights=[0.8, 0.2])[
                0
            ]
        if "deep" in method:
            max_steps = random.choices(
                [40, 80, 100, 1000, 5000, 10000, 50000],
                [0.2, 0.2, 0.2, 0.1, 0.05, 0.05, 0.01],
            )[0]
        else:
            max_steps = random.choices(
                [40, 80, 100, 1000, 5000],
                [0.2, 0.2, 0.2, 0.05, 0.03],
            )[0]
        activation = random.choices(
            ['ReLU', 'Softplus', 'Tanh', 'SELU', 'LeakyReLU', 'PReLU', 'Sigmoid'],
            [0.5, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
        )[0]
        loss = random.choices(
            [
                'MQLoss',
                'HuberMQLoss',
                'Poisson',
                'Bernoulli',
                'NegativeBinomial',
                'Normal',
                'Tweedie',
                'HuberLoss',
                "MAE",
                "SMAPE",
                "StudentT",
            ],
            [0.5, 0.1, 0.1, 0.01, 0.1, 0.1, 0.01, 0.1, 0.1, 0.1, 0.01],
        )[0]
        point_quantile = None
        if loss in ["MQLoss", "HuberMQLoss"]:
            point_quantile = random.choices(
                [None, 0.35, 0.45, 0.55, 0.65, 0.7], [0.5, 0.1, 0.1, 0.1, 0.1, 0.1]
            )[0]
        if models == "TFT":
            model_args = {
                "n_head": random.choice([2, 4]),
                "dropout": random.choice([0.05, 0.2, 0.3, 0.6]),
                "batch_size": random.choice([32, 64, 128, 256]),
                "hidden_size": random.choice([4, 64, 128, 256]),
                "windows_batch_size": random.choice([128, 256, 512, 1024]),
            }
        elif models == "PatchTST":
            model_args = {
                "n_heads": random.choice([2, 4, 16, 32]),
                "dropout": random.choice([0.05, 0.2, 0.3, 0.6]),
                "batch_size": random.choice([32, 64, 128, 256]),
                "hidden_size": random.choice([4, 64, 128, 256]),
                "linear_hidden_size": random.choice([4, 64, 128, 256, 512, 1024]),
                "windows_batch_size": random.choice([128, 256, 512, 1024]),
                'encoder_layers': random.choices([1, 2, 3, 4], [0.2, 0.4, 0.6, 0.1])[0],
            }
        elif models == "NHITS":
            model_args = {
                "input_size": random.choice([28, 28 * 2, 28 * 3, 28 * 5]),
                "n_blocks": 5 * [1],
                "mlp_units": 5 * [[512, 512]],
                "n_pool_kernel_size": random.choice(
                    [5 * [1], 5 * [2], 5 * [4], [8, 4, 2, 1, 1]]
                ),
                "n_freq_downsample": random.choice([[8, 4, 2, 1, 1], [1, 1, 1, 1, 1]]),
                "batch_size": random.choice([32, 64, 128, 256]),
                "windows_batch_size": random.choice([128, 256, 512, 1024]),
                "activation": activation,
            }
        elif models == "MLP":
            model_args = {
                'num_layers': random.choices([1, 2, 3, 4], [0.6, 0.4, 0.2, 0.1])[0],
                'hidden_size': random.choice([1024, 512, 2048, 256, 2560, 3072, 4096]),
            }
        elif models == "TiDE":
            model_args = {
                # borrowed mostly from the defaults used for the paper results
                "layernorm": random.choices([True, False], [0.2, 0.8])[0],
                "dropout": random.choices([0.0, 0.3, 0.5, 0.7], [0.3, 0.3, 0.3, 0.05])[
                    0
                ],
                "batch_size": random.choices(
                    [1024, 512, 257, 32], [0.05, 0.4, 0.1, 0.1]
                )[0],
                "hidden_size": random.choices(
                    [1024, 512, 256, 64], [0.1, 0.2, 0.4, 0.1]
                )[0],
                "num_encoder_layers": random.choices([1, 2, 3], [0.85, 0.1, 0.02])[0],
                "num_decoder_layers": random.choices([1, 2, 3], [0.85, 0.1, 0.02])[0],
                "decoder_output_dim": random.choices(
                    [16, 8, 4, 32], [0.4, 0.4, 0.4, 0.1]
                )[0],
                "temporal_decoder_dim": random.choices(
                    [128, 64, 32, 16], [0.1, 0.8, 0.1, 0.1]
                )[0],
            }
        else:
            model_args = {}

        return {
            'model': models,
            'scaler_type': random.choices(
                ["identity", 'robust', 'minmax', 'standard'], [0.5, 0.5, 0.2, 0.2]
            )[0],
            'loss': loss,
            'learning_rate': random.choices(
                [0.001, 0.1, 0.01, 0.0003, 0.00001, 0.000001],
                [0.4, 0.1, 0.1, 0.1, 0.1, 0.02],
            )[0],
            "max_steps": max_steps,
            'input_size': random.choices(
                [10, 28, "2ForecastLength", "3ForecastLength"], [0.2, 0.2, 0.2, 0.2]
            )[0],
            # "early_stop_patience_steps": random.choice([1, 3, 5]),
            "point_quantile": point_quantile,
            "model_args": model_args,
            'regression_type': regression_type_choice,
        }

    def get_params(self):
        """Return dict of current parameters."""
        return {
            # NOTE: `models` (plural) conflicts with AutoTS ensemble setup
            'model': self.model,
            'scaler_type': self.scaler_type,
            'loss': self.loss,
            'learning_rate': self.learning_rate,
            "max_steps": self.max_steps,
            'input_size': self.input_size,
            'point_quantile': self.point_quantile,
            "model_args": self.model_args,
            'regression_type': self.regression_type,
        }


if False:
    # from autots.models.neural_forecast import NeuralForecast
    from autots import load_daily, create_regressor, infer_frequency

    df = load_daily(long=False)
    forecast_length = 28
    frequency = infer_frequency(df)
    regr_train, regr_fcst = create_regressor(
        df,
        forecast_length=forecast_length,
        frequency=frequency,
        drop_most_recent=0,
        scale=True,
        summarize="auto",
        backfill="bfill",
        fill_na="pchip",
        holiday_countries=["US"],
        datepart_method="recurring",
        preprocessing_params={
            "fillna": None,
            "transformations": {"0": "LocalLinearTrend"},
            "transformation_params": {
                "0": {
                    'rolling_window': 30,
                    'n_tails': 0.1,
                    'n_future': 0.2,
                    'method': 'mean',
                    'macro_micro': True,
                },
            },
        },
    )

    params = NeuralForecast().get_new_params()
    print(params)
    mod = NeuralForecast(forecast_length=forecast_length, frequency=frequency, **params)
    mod.fit(df, future_regressor=regr_train)
    prediction = mod.predict(future_regressor=regr_fcst)
    prediction.plot_grid(df)



# -*- coding: utf-8 -*-
"""Overall testing."""
import unittest
import numpy as np
import pandas as pd
from autots.datasets import (
    load_daily, load_monthly, load_artificial, load_sine
)
from autots.tools.transform import ThetaTransformer, FIRFilter

class TestTransforms(unittest.TestCase):
    
    def test_theta(self):
        # Sample DataFrame with a DatetimeIndex and multiple time series columns
        dates = pd.date_range(start='2020-01-01', periods=100, freq='D')
        data = pd.DataFrame({
            'series1': np.random.randn(100).cumsum(),
            'series2': np.random.randn(100).cumsum(),
        }, index=dates)

        theta_values = [0, 1, 2]
        theta_transformer = ThetaTransformer(theta_values=theta_values)
        
        params = theta_transformer.get_new_params()
        self.assertTrue(params)

        theta_transformer.fit(data)
        transformed_data = theta_transformer.transform(data)
        reconstructed_data = theta_transformer.inverse_transform(transformed_data)
        self.assertTrue(np.allclose(data.values, reconstructed_data.values, atol=1e-8))

    def test_firfilter(self):
        df = load_daily(long=False)
        transformer = FIRFilter()
        transformed = transformer.fit_transform(df)
        inverse = transformer.inverse_transform(transformed)  # noqa
        
        if False:
            col = df.columns[0]
            pd.concat([df[col], transformed[col].rename("transformed")], axis=1).plot()
        
        self.assertCountEqual(transformed.index.tolist(), df.index.tolist())
        self.assertCountEqual(transformed.columns.tolist(), df.columns.tolist())
